import json as JSON
from confluent_kafka import Producer
from .output import _OUTPUT_GRAPHVIZSHAPE
from ..base import Node


def Kafka(node, servers='', topic='', json=False, wrap=False):
    '''Connect to kafka server and send data

    Args:
        foo (callable): input stream
        foo_kwargs (dict): kwargs for the input stream
        servers (list): kafka bootstrap servers
        group (str): kafka group id
        topics (list): list of kafka topics to connect to
        json (bool): load input data as json
        wrap (bool): wrap result in a list
        interval (int): kafka poll interval
    '''

    p = Producer({'bootstrap.servers': servers})

    async def _send(data, producer=p, topic=topic, json=json, wrap=wrap):
        ret = []
        # Trigger any available delivery report callbacks from previous produce() calls
        producer.poll(0)

        if wrap:
            data = [data]

        if json:
            data = JSON.dumps(data)

        producer.produce(topic, data.encode('utf-8'), callback=lambda *args: ret.append(args))

        for data in ret:
            yield data
        ret = []

    ret = Node(foo=_send, name='Kafka', inputs=1, graphvizshape=_OUTPUT_GRAPHVIZSHAPE)
    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from enum import Enum


class Plugin(Enum):
    '''The plugins (grids/charts) available in Perspective.  Pass these into
    the `plugin` arg in `PerspectiveWidget` or `PerspectiveViewer`.

    Examples:
        >>> widget = PerspectiveWidget(data, plugin=Plugin.TREEMAP)
    '''
    HYPERGRID = 'hypergrid'  # hypergrid
    GRID = 'hypergrid'  # hypergrid

    YBAR = 'y_bar'  # highcharts
    XBAR = 'x_bar'  # highcharts
    YLINE = 'y_line'  # highcharts
    YAREA = 'y_area'  # highcharts
    YSCATTER = 'y_scatter'  # highcharts
    XYLINE = 'xy_line'  # highcharts
    XYSCATTER = 'xy_scatter'  # highcharts
    TREEMAP = 'treemap'  # highcharts
    SUNBURST = 'sunburst'  # highcharts
    HEATMAP = 'heatmap'  # highcharts

    YBAR_D3 = 'd3_y_bar'  # d3fc
    XBAR_D3 = 'd3_x_bar'  # d3fc
    YLINE_D3 = 'd3_y_line'  # d3fc
    YAREA_D3 = 'd3_y_area'  # d3fc
    YSCATTER_D3 = 'd3_y_scatter'  # d3fc
    XYSCATTER_D3 = 'd3_xy_scatter'  # d3fc
    TREEMAP_D3 = 'd3_treemap'  # d3fc
    SUNBURST_D3 = 'd3_sunburst'  # d3fc
    HEATMAP_D3 = 'd3_heatmap'  # d3fc

    CANDLESTICK = 'd3_candlestick'  # d3fc
    CANDLESTICK_D3 = 'd3_candlestick'  # d3fc
    OHLC = 'd3_ohlc'  # d3fc
    OHLC_D3 = 'd3_ohlc'  # d3fc

    @staticmethod
    def options():
        return list(c.value for c in Plugin)



# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import six
import numpy as np
from pytest import raises
from perspective import PerspectiveError
from perspective.table import Table
from datetime import date, datetime


class TestTableNumpy(object):
    def test_empty_table(self):
        tbl = Table([])
        assert tbl.size() == 0

    def test_table_int(self):
        data = {"a": np.array([1, 2, 3]), "b": np.array([4, 5, 6])}
        tbl = Table(data)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3],
            "b": [4, 5, 6]
        }

    def test_table_int_lots_of_columns(self):
        data = {
            "a": np.array([1, 2, 3]),
            "b": np.array([4, 5, 6]),
            "c": np.array([4, 5, 6]),
            "d": np.array([4, 5, 6]),
            "e": np.array([4, 5, 6]),
            "f": np.array([4, 5, 6]),
        }
        tbl = Table(data)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3],
            "b": [4, 5, 6],
            "c": [4, 5, 6],
            "d": [4, 5, 6],
            "e": [4, 5, 6],
            "f": [4, 5, 6]
        }

    def test_table_int_with_None(self):
        data = {"a": np.array([1, 2, 3, None, None]), "b": np.array([4, 5, 6, None, None])}
        tbl = Table(data)
        assert tbl.size() == 5
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3, None, None],
            "b": [4, 5, 6, None, None]
        }

    def test_table_int8(self):
        data = {"a": np.array([1, 2, 3]).astype(np.int8), "b": np.array([4, 5, 6]).astype(np.int8)}
        tbl = Table(data)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3],
            "b": [4, 5, 6]
        }

    def test_table_int16(self):
        data = {"a": np.array([1, 2, 3]).astype(np.int16), "b": np.array([4, 5, 6]).astype(np.int16)}
        tbl = Table(data)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3],
            "b": [4, 5, 6]
        }

    def test_table_int32(self):
        data = {"a": np.array([1, 2, 3]).astype(np.int32), "b": np.array([4, 5, 6]).astype(np.int32)}
        tbl = Table(data)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3],
            "b": [4, 5, 6]
        }

    def test_table_int64(self):
        data = {"a": np.array([1, 2, 3]).astype(np.int64), "b": np.array([4, 5, 6]).astype(np.int64)}
        tbl = Table(data)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3],
            "b": [4, 5, 6]
        }

    def test_table_long_numpy(self):
        if six.PY2:
            data = {"a": np.array([1, 2, 3], dtype=long)}  # noqa: F821
            tbl = Table(data)
            assert tbl.schema() == {
                "a": int
            }
            assert tbl.size() == 3
            assert tbl.view().to_dict() == {
                "a": [1, 2, 3]
            }

    def test_table_float(self):
        data = {"a": np.array([1.1, 2.2]), "b": np.array([3.3, 4.4])}
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "a": [1.1, 2.2],
            "b": [3.3, 4.4]
        }

    def test_table_float32(self):
        data = {"a": np.array([1.1, 2.2]).astype(np.float32), "b": np.array([3.3, 4.4]).astype(np.float32)}
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            # py::cast automatically upcasts to 64-bit float
            "a": [1.100000023841858, 2.200000047683716],
            "b": [3.299999952316284, 4.400000095367432]
        }

    def test_table_float64(self):
        data = {"a": np.array([1.1, 2.2]).astype(np.float64), "b": np.array([3.3, 4.4]).astype(np.float64)}
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "a": [1.1, 2.2],
            "b": [3.3, 4.4]
        }

    # booleans

    def test_table_bool(self):
        data = {"a": np.array([True, False]), "b": np.array([False, True])}
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "a": [True, False],
            "b": [False, True]
        }

    def test_table_bool8(self):
        data = {"a": np.array([True, False]).astype(np.bool8), "b": np.array([False, True]).astype(np.bool8)}
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "a": [True, False],
            "b": [False, True]
        }

    def test_table_bool_with_none(self):
        data = {"a": np.array([True, False, None, False]), "b": np.array([False, True, None, False])}
        tbl = Table(data)
        assert tbl.size() == 4
        assert tbl.view().to_dict() == {
            "a": [True, False, None, False],
            "b": [False, True, None, False]
        }

    def test_table_bool_with_dtype(self):
        data = {"a": np.array([True, False, False], dtype="?"), "b": np.array([False, True, False], dtype="?")}
        tbl = Table(data)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "a": [True, False, False],
            "b": [False, True, False]
        }

    def test_table_bool_str(self):
        data = {"a": np.array(["True", "False"]), "b": np.array(["False", "True"])}
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.schema() == {
            "a": bool,
            "b": bool
        }
        assert tbl.view().to_dict() == {
            "a": [True, False],
            "b": [False, True]
        }    

    # strings

    def test_table_str_object(self):
        data = {"a": np.array(["abc", "def"], dtype=object), "b": np.array(["hij", "klm"], dtype=object)}
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "a": ["abc", "def"],
            "b": ["hij", "klm"]
        }

    def test_table_str_dtype(self):
        dtype = "U3"
        if six.PY2:
            dtype = "|S3"
        data = {"a": np.array(["abc", "def"], dtype=dtype), "b": np.array(["hij", "klm"], dtype=dtype)}
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "a": ["abc", "def"],
            "b": ["hij", "klm"]
        }

    def test_table_unicode_py2(self):
        if six.PY2:
            data = {
                "a": np.array([unicode("abc"), unicode("def")]),  # noqa: F821
                "b": np.array([unicode("hij"), unicode("klm")])  # noqa: F821
            }
            tbl = Table(data)
            assert tbl.size() == 2
            assert tbl.view().to_dict() == {
                "a": ["abc", "def"],
                "b": ["hij", "klm"]
            }

    # date and datetime

    def test_table_date(self):
        data = {"a": np.array([date(2019, 7, 11)]), "b": np.array([date(2019, 7, 12)])}
        tbl = Table(data)
        assert tbl.size() == 1
        assert tbl.schema() == {
            "a": date,
            "b": date
        }
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 11)],
            "b": [datetime(2019, 7, 12)]
        }

    def test_table_np_datetime(self):
        data = {"a": np.array([datetime(2019, 7, 11, 12, 13)], dtype=np.datetime64), "b": np.array([datetime(2019, 7, 11, 12, 14)], dtype=np.datetime64)}
        tbl = Table(data)
        assert tbl.size() == 1
        assert tbl.schema() == {
            "a": datetime,
            "b": datetime
        }
        assert tbl.view().to_numpy() == data

    def test_table_np_datetime_mixed_dtype(self):
        data = {"a": np.array([datetime(2019, 7, 11, 12, 13)], dtype=np.datetime64), "b": np.array([datetime(2019, 7, 11, 12, 14)], dtype=object)}
        tbl = Table(data)
        assert tbl.size() == 1
        assert tbl.schema() == {
            "a": datetime,
            "b": datetime
        }
        assert tbl.view().to_numpy() == data

    def test_table_np_datetime_default(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype=np.datetime64)
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0)]
        }

    def test_table_np_datetime_string_dtype(self):
        data = ["2019/07/11 15:30:05", "2019/07/11 15:30:05"]
        tbl = Table({
            "a": np.array(data)
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 11, 15, 30, 5), datetime(2019, 7, 11, 15, 30, 5)]
        }

    def test_table_np_datetime_string_on_schema(self):
        data = ["2019/07/11 15:30:05", "2019/07/11 15:30:05"]
        tbl = Table({
            "a": datetime
        })

        tbl.update({"a": data})

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 11, 15, 30, 5), datetime(2019, 7, 11, 15, 30, 5)]
        }

    def test_table_np_datetime_ns(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[ns]")
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0)]
        }

    def test_table_np_datetime_us(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[us]")
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0)]
        }

    def test_table_np_datetime_ms(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[ms]")
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0)]
        }

    def test_table_np_datetime_s(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[s]")
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0)]
        }

    def test_table_np_datetime_m(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[m]")
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0)]
        }

    def test_table_np_datetime_h(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[h]")
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0)]
        }

    def test_table_np_datetime_D(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[D]")
        })

        assert tbl.schema() == {
            "a": date
        }

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 0, 0)]
        }

    def test_table_np_datetime_W(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[W]")
        })

        assert tbl.schema() == {
            "a": date
        }

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 11, 0, 0)]
        }

    def test_table_np_datetime_M(self):
        tbl = Table({
            "a": np.array([
                datetime(2019, 5, 12, 11, 0),
                datetime(2019, 6, 12, 11, 0),
                datetime(2019, 7, 12, 11, 0)],
                dtype="datetime64[M]")
        })

        assert tbl.schema() == {
            "a": date
        }

        assert tbl.view().to_dict() == {
            "a": [
                datetime(2019, 5, 1, 0, 0),
                datetime(2019, 6, 1, 0, 0),
                datetime(2019, 7, 1, 0, 0)
            ]
        }

    def test_table_np_datetime_Y(self):
        tbl = Table({
            "a": np.array([
                datetime(2017, 5, 12, 11, 0),
                datetime(2018, 6, 12, 11, 0),
                datetime(2019, 7, 12, 11, 0)],
                dtype="datetime64[Y]")
        })

        assert tbl.schema() == {
            "a": date
        }

        assert tbl.view().to_dict() == {
            "a": [
                datetime(2017, 1, 1, 0, 0),
                datetime(2018, 1, 1, 0, 0),
                datetime(2019, 1, 1, 0, 0)
            ]
        }

    def test_table_np_datetime_ms_nat(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0), np.datetime64("nat")], dtype="datetime64[ms]")
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0), None]
        }

    def test_table_np_datetime_s_nat(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0), np.datetime64("nat")], dtype="datetime64[s]")
        })

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0), None]
        }

    def test_table_np_timedelta(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[ns]") - np.array([datetime(2019, 7, 1, 11, 0)], dtype="datetime64[ns]")
        })

        assert tbl.schema() == {
            "a": str
        }

        assert tbl.view().to_dict() == {
            "a": ["950400000000000 nanoseconds"]
        }

    def test_table_np_timedelta_us(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[us]") - np.array([datetime(2019, 7, 1, 11, 0)], dtype="datetime64[us]")
        })

        assert tbl.schema() == {
            "a": str
        }

        assert tbl.view().to_dict() == {
            "a": ["950400000000 microseconds"]
        }

    def test_table_np_timedelta_ms(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[ms]") - np.array([datetime(2019, 7, 1, 11, 0)], dtype="datetime64[ms]")
        })

        assert tbl.schema() == {
            "a": str
        }

        assert tbl.view().to_dict() == {
            "a": ["950400000 milliseconds"]
        }

    def test_table_np_timedelta_s(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[s]") - np.array([datetime(2019, 7, 1, 11, 0)], dtype="datetime64[s]")
        })

        assert tbl.schema() == {
            "a": str
        }

        assert tbl.view().to_dict() == {
            "a": ["950400 seconds"]
        }

    def test_table_np_timedelta_m(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[m]") - np.array([datetime(2019, 7, 1, 11, 0)], dtype="datetime64[m]")
        })

        assert tbl.schema() == {
            "a": str
        }

        assert tbl.view().to_dict() == {
            "a": ["15840 minutes"]
        }

    def test_table_np_timedelta_h(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[h]") - np.array([datetime(2019, 7, 1, 11, 0)], dtype="datetime64[h]")
        })

        assert tbl.schema() == {
            "a": str
        }

        assert tbl.view().to_dict() == {
            "a": ["264 hours"]
        }

    def test_table_np_timedelta_d(self):
        tbl = Table({
            "a": np.array([datetime(2019, 7, 12, 11, 0)], dtype="datetime64[D]") - np.array([datetime(2019, 7, 1, 11, 0)], dtype="datetime64[D]")
        })

        assert tbl.schema() == {
            "a": str
        }

        assert tbl.view().to_dict() == {
            "a": ["11 days"]
        }

    def test_table_np_timedelta_with_none(self):
        tbl = Table({
            "a": np.array([None, datetime(2019, 7, 12, 11, 0)], dtype="datetime64[ns]") - np.array([datetime(2019, 7, 1, 11, 0), None], dtype="datetime64[ns]")
        })

        assert tbl.schema() == {
            "a": str
        }

        assert tbl.view().to_dict() == {
            "a": [None, None]  # two `NaT` values
        }

    def test_table_np_mixed(self):
        data = {
            "a": np.arange(5),
            "b": np.full(5, np.nan),
            "c": ["a", "b", "c", "d", "e"]
        }

        # should not be able to parse mixed dicts of numpy array with list
        with raises(PerspectiveError):
            Table(data)

    def test_table_np_promote(self):
        data = {
            "a": np.arange(5),
            "b": np.full(5, np.nan),
            "c": np.array([1, 2, 3, 2147483648, 5])
        }
        tbl = Table({
            "a": int,
            "b": float,
            "c": int
        })
        tbl.update(data)
        assert tbl.size() == 5
        assert tbl.schema() == {
            "a": int,
            "b": float,
            "c": int
        }

        assert tbl.view().to_dict() == {
            "a": [0, 1, 2, 3, 4],
            "b": [None, None, None, None, None],
            "c": [1.0, 2.0, 3.0, 2147483648.0, 5.0]
        }

    def test_table_np_promote_to_string(self):
        data = {
            "a": np.arange(4),
            "b": np.array([1, 2, "abc", "abc"]),
        }
        tbl = Table(data)
        assert tbl.size() == 4
        assert tbl.schema() == {
            "a": int,
            "b": str,
        }

        assert tbl.view().to_dict() == {
            "a": [0, 1, 2, 3],
            "b": ["1", "2", "abc", "abc"],
        }

    def test_table_np_implicit_index(self):
        data = {
            "a": np.array(["a", "b", "c", "d", "e"]),
            "b": np.array([1, 2, 3, 4, 5])
        }
        tbl = Table(data)
        assert tbl.size() == 5
        assert tbl.schema() == {
            "a": str,
            "b": int
        }
        tbl.update({
            "__INDEX__": np.array([1, 2, 3, 4]),
            "a": np.array(["bb", "cc", "dd", "ee"])
        })

        assert tbl.view().to_dict() == {
            "a": ["a", "bb", "cc", "dd", "ee"],
            "b": [1, 2, 3, 4, 5]
        }

    # from schema

    def test_table_numpy_from_schema_int(self):
        df = {
            "a": np.array([1, None, 2, None, 3, 4])
        }
        table = Table({
            "a": int
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [1, None, 2, None, 3, 4]

    def test_table_numpy_from_schema_long(self):
        if six.PY2:
            df = {
                "a": np.array([1, None, 2, None, 3, 4])
            }
            table = Table({
                "a": long  # noqa: F821
            })
            table.update(df)
            assert table.view().to_dict()["a"] == [1, None, 2, None, 3, 4]

    def test_table_numpy_from_schema_int_to_long(self):
        if six.PY2:
            df = {
                "a": np.array([1, 2, 3, 4], dtype="int64")
            }
            table = Table({
                "a": long  # noqa: F821
            })
            table.update(df)
            assert table.view().to_dict()["a"] == [1, 2, 3, 4]

    def test_table_numpy_from_schema_float_to_long(self):
        if six.PY2:
            df = {
                "a": np.array([1, None, 2, None, 3, 4], dtype="float64")
            }
            table = Table({
                "a": long  # noqa: F821
            })
            table.update(df)
            assert table.view().to_dict()["a"] == [1, None, 2, None, 3, 4]

    def test_table_numpy_from_schema_bool(self):
        data = [True, False, True, False]
        df = {
            "a": data
        }
        table = Table({
            "a": bool
        })
        table.update(df)
        assert table.view().to_dict()["a"] == data

    def test_table_numpy_from_schema_float(self):
        data = [1.5, None, 2.5, None, 3.5, 4.5]
        df = {"a": np.array(data)}
        table = Table({
            "a": float
        })
        table.update(df)
        assert table.view().to_dict()["a"] == data

    def test_table_numpy_from_schema_float_all_nan(self):
        data = [np.nan, np.nan, np.nan, np.nan]
        df = {"a": np.array(data)}
        table = Table({
            "a": float
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [None, None, None, None]

    def test_table_numpy_from_schema_float_to_int(self):
        data = [None, 1.5, None, 2.5, None, 3.5, 4.5]
        df = {"a": np.array(data)}
        table = Table({
            "a": int
        })
        table.update(df)
        # truncates decimal
        assert table.view().to_dict()["a"] == [None, 1, None, 2, None, 3, 4]

    def test_table_numpy_from_schema_int_to_float(self):
        data = [None, 1, None, 2, None, 3, 4]
        df = {"a": np.array(data)}
        table = Table({
            "a": float
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [None, 1.0, None, 2.0, None, 3.0, 4.0]

    def test_table_numpy_from_schema_date(self):
        data = [date(2019, 8, 15), None, date(2019, 8, 16)]
        df = {"a": np.array(data)}
        table = Table({
            "a": date
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [datetime(2019, 8, 15), None, datetime(2019, 8, 16)]

    def test_table_numpy_from_schema_datetime(self):
        data = [datetime(2019, 7, 11, 12, 30, 5), None, datetime(2019, 7, 11, 13, 30, 5), None]
        df = {"a": np.array(data)}
        table = Table({
            "a": datetime
        })
        table.update(df)
        assert table.view().to_dict()["a"] == data

    def test_table_numpy_from_schema_datetime_timestamp_s(self, util):
        data = [util.to_timestamp(datetime(2019, 7, 11, 12, 30, 5)), np.nan, util.to_timestamp(datetime(2019, 7, 11, 13, 30, 5)), np.nan]
        df = {"a": np.array(data)}
        table = Table({
            "a": datetime
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [
            datetime(2019, 7, 11, 12, 30, 5),
            None,
            datetime(2019, 7, 11, 13, 30, 5),
            None
        ]

    def test_table_numpy_from_schema_datetime_timestamp_ms(self, util):
        data = [
            util.to_timestamp(datetime(2019, 7, 11, 12, 30, 5)) * 1000,
            np.nan,
            util.to_timestamp(datetime(2019, 7, 11, 13, 30, 5)) * 1000,
            np.nan
        ]

        df = {"a": np.array(data)}
        table = Table({
            "a": datetime
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [
            datetime(2019, 7, 11, 12, 30, 5),
            None,
            datetime(2019, 7, 11, 13, 30, 5),
            None
        ]

    def test_table_numpy_from_schema_str(self):
        data = ["a", None, "b", None, "c"]
        df = {"a": np.array(data)}
        table = Table({
            "a": str
        })
        table.update(df)
        assert table.view().to_dict()["a"] == data

    # partial update

    def test_table_numpy_partial_update(self):
        data = ["a", None, "b", None, "c"]
        df = {"a": np.array([1, 2, 3, 4, 5]), "b": np.array(data), "c": np.array(data)}
        table = Table(df, index="a")
        table.update({
            "a": np.array([2, 4, 5]),
            "b": np.array(["x", "y", "z"])
        })
        assert table.view().to_dict() == {
            "a": [1, 2, 3, 4, 5],
            "b": ["a", "x", "b", "y", "z"],
            "c": ["a", None, "b", None, "c"]
        }

    def test_table_numpy_partial_update_implicit(self):
        data = ["a", None, "b", None, "c"]
        df = {"a": np.array([1, 2, 3, 4, 5]), "b": np.array(data), "c": np.array(data)}
        table = Table(df)
        table.update({
            "__INDEX__": np.array([1, 3, 4]),
            "b": np.array(["x", "y", "z"])
        })
        assert table.view().to_dict() == {
            "a": [1, 2, 3, 4, 5],
            "b": ["a", "x", "b", "y", "z"],
            "c": ["a", None, "b", None, "c"]
        }

    # structured array

    def test_table_structured_array(self):
        d = np.array([(1.0, 2), (3.0, 4)], dtype=[('x', '<f8'), ('y', '<i8')])
        table = Table(d)
        assert table.schema() == {
            "x": float,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [1.0, 3.0],
            "y": [2, 4]
        }

    # recarray

    def test_table_recarray(self):
        d = np.array([(1.0, 2), (3.0, 4)], dtype=[('x', '<f8'), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": float,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [1.0, 3.0],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_ns(self):
        d = np.array([(datetime(2019, 7, 11, 8, 30, 29), 2), (datetime(2019, 7, 11, 8, 30, 29), 4)], dtype=[('x', "datetime64[ns]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": datetime,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [datetime(2019, 7, 11, 8, 30, 29), datetime(2019, 7, 11, 8, 30, 29)],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_us(self):
        d = np.array([(datetime(2019, 7, 11, 8, 30, 29), 2), (datetime(2019, 7, 11, 8, 30, 29), 4)], dtype=[('x', "datetime64[us]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": datetime,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [datetime(2019, 7, 11, 8, 30, 29), datetime(2019, 7, 11, 8, 30, 29)],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_ms(self):
        d = np.array([(datetime(2019, 7, 11, 8, 30, 29), 2), (datetime(2019, 7, 11, 8, 30, 29), 4)], dtype=[('x', "datetime64[ms]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": datetime,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [datetime(2019, 7, 11, 8, 30, 29), datetime(2019, 7, 11, 8, 30, 29)],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_s(self):
        d = np.array([(datetime(2019, 7, 11, 8, 30, 29), 2), (datetime(2019, 7, 11, 8, 30, 29), 4)], dtype=[('x', "datetime64[s]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": datetime,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [datetime(2019, 7, 11, 8, 30, 29), datetime(2019, 7, 11, 8, 30, 29)],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_m(self):
        d = np.array([(datetime(2019, 7, 11, 8, 30, 29), 2), (datetime(2019, 7, 11, 8, 31, 29), 4)], dtype=[('x', "datetime64[m]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": datetime,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [datetime(2019, 7, 11, 8, 30, 0), datetime(2019, 7, 11, 8, 31, 0)],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_h(self):
        d = np.array([(datetime(2019, 7, 11, 8, 30, 29), 2), (datetime(2019, 7, 11, 9, 30, 29), 4)], dtype=[('x', "datetime64[h]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": datetime,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [datetime(2019, 7, 11, 8, 0, 0), datetime(2019, 7, 11, 9, 0, 0)],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_D(self):
        d = np.array([(datetime(2019, 7, 11, 8, 30, 29), 2), (datetime(2019, 7, 12, 8, 30, 29), 4)], dtype=[('x', "datetime64[D]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": date,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [datetime(2019, 7, 11, 0, 0, 0), datetime(2019, 7, 12, 0, 0, 0)],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_W(self):
        d = np.array([(datetime(2019, 6, 30, 0, 0, 0), 2), (datetime(2019, 7, 7, 0, 0, 0), 4)], dtype=[('x', "datetime64[W]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": date,
            "y": int
        }
        assert table.view().to_dict() == {
            # one week apart
            "x": [datetime(2019, 6, 27, 0, 0, 0), datetime(2019, 7, 4, 0, 0, 0)],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_M(self):
        d = np.array([(datetime(2019, 6, 11, 8, 30, 29), 2), (datetime(2019, 7, 11, 8, 30, 29), 4)], dtype=[('x', "datetime64[M]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": date,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [datetime(2019, 6, 1, 0, 0, 0), datetime(2019, 7, 1, 0, 0, 0)],
            "y": [2, 4]
        }

    def test_table_recarray_datetime_Y(self):
        d = np.array([(datetime(2018, 7, 11, 8, 30, 29), 2), (datetime(2019, 7, 11, 8, 30, 29), 4)], dtype=[('x', "datetime64[Y]"), ('y', '<i8')]).view(np.recarray)
        table = Table(d)
        assert table.schema() == {
            "x": date,
            "y": int
        }
        assert table.view().to_dict() == {
            "x": [datetime(2018, 1, 1, 0, 0, 0), datetime(2019, 1, 1, 0, 0, 0)],
            "y": [2, 4]
        }

    def test_table_recarray_str(self):
        table = Table(np.array([("string1", "string2"), ("string3", "string4")], dtype=[('x', 'O'), ('y', 'O')]).view(np.recarray))
        assert table.schema() == {
            "x": str,
            "y": str
        }
        assert table.view().to_dict() == {
            "x": ["string1", "string3"],
            "y": ["string2", "string4"]
        }

    def test_table_recarray_str_dtype(self):
        dtype = "U7"
        if six.PY2:
            dtype = "|S7"
        table = Table(np.array([("string1", "string2"), ("string3", "string4")], dtype=[('x', dtype), ('y', dtype)]).view(np.recarray))
        assert table.schema() == {
            "x": str,
            "y": str
        }
        assert table.view().to_dict() == {
            "x": ["string1", "string3"],
            "y": ["string2", "string4"]
        }



import requests
import json as JSON
from .output import _OUTPUT_GRAPHVIZSHAPE
from ..base import Node
from ...base import StreamEnd


def HTTP(node, url='', json=False, wrap=False, field=None, proxies=None, cookies=None):
    '''Connect to url and post results to it

    Args:
        foo (callable): input stream
        foo_kwargs (dict): kwargs for the input stream
        url (str): url to post to
        json (bool): dump data as json
        wrap (bool): wrap input in a list
        field (str): field to index result by
        proxies (list): list of URL proxies to pass to requests.post
        cookies (list): list of cookies to pass to requests.post
    '''

    def _send(data, url=url, json=json, wrap=wrap, field=field, proxies=proxies, cookies=cookies):
        if wrap:
            data = [data]
        if json:
            data = JSON.dumps(data)

        msg = requests.post(url, data=data, cookies=cookies, proxies=proxies)

        if msg is None:
            return StreamEnd()

        if msg.status_code != 200:
            return msg

        if json:
            msg = msg.json()

        if field:
            msg = msg[field]

        if wrap:
            msg = [msg]

        return msg

    ret = Node(foo=_send, name='Http', inputs=1, graphvizshape=_OUTPUT_GRAPHVIZSHAPE)
    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret



from .input import *  # noqa: F401, F403
from .file import File, File as FileSource  # noqa: F401
from .http import HTTP, HTTP as HTTPSource  # noqa: F401
from .kafka import Kafka, Kafka as KafkaSource  # noqa: F401
from .socketio import SocketIO, SocketIO as SocketIOSource  # noqa: F401
from .ws import WebSocket, WebSocket as WebSocketSource  # noqa: F401



import tributary.streaming as ts


def foo():
    yield 1
    yield 1
    yield 1
    yield 1
    yield 1


def foo2():
    yield 1
    yield 2
    yield 0
    yield 5
    yield 4


def foo3():
    yield 1
    yield 2
    yield 3
    yield 4
    yield 5


class TestRolling:
    def test_count(self):
        assert ts.run(ts.RollingCount(ts.Foo(foo))) == [1, 2, 3, 4, 5]

    def test_sum(self):
        assert ts.run(ts.RollingSum(ts.Foo(foo))) == [1, 2, 3, 4, 5]

    def test_min(self):
        assert ts.run(ts.RollingMin(ts.Foo(foo2))) == [1, 1, 0, 0, 0]

    def test_max(self):
        assert ts.run(ts.RollingMax(ts.Foo(foo2))) == [1, 2, 2, 5, 5]

    def test_average(self):
        assert ts.run(ts.RollingAverage(ts.Foo(foo3))) == [1, 1.5, 2, 2.5, 3]



import json as JSON
import websockets

from .input import Foo
from ...base import StreamNone, StreamEnd


class WebSocket(Foo):
    '''Connect to websocket and yield back results

    Args:
        url (str): websocket url to connect to
        json (bool): load websocket data as json
        wrap (bool): wrap result in a list
    '''

    def __init__(self, url, json=False, wrap=False):
        async def _listen(url=url, json=json, wrap=wrap):
            async with websockets.connect(url) as websocket:
                async for x in websocket:
                    if isinstance(x, StreamNone):
                        continue
                    elif not x or isinstance(x, StreamEnd):
                        break

                    if json:
                        x = JSON.loads(x)
                    if wrap:
                        x = [x]
                    yield x

        super().__init__(foo=_listen)
        self._name = 'WebSocket'



import asyncio
from copy import deepcopy
from .base import StreamingGraph, Node as StreamingNode  # noqa: F401
from .calculations import *  # noqa: F401, F403
from .input import *  # noqa: F401, F403
from .output import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403
from ..base import StreamEnd, StreamNone, StreamRepeat


async def _run(node):
    ret = []
    nodes = node._deep_bfs()
    while True:
        for level in nodes:
            await asyncio.gather(*(asyncio.create_task(n()) for n in level))
        val = deepcopy(node.value())
        if not isinstance(val, (StreamEnd, StreamNone, StreamRepeat)):
            ret.append(val)
        elif isinstance(val, StreamEnd):
            break
    return ret


def run(node):
    loop = asyncio.get_event_loop()
    if loop.is_running():
        # return future
        return asyncio.create_task(_run(node))
    # block until done
    return loop.run_until_complete(_run(node))



# coding: utf-8
from .widget import DagreD3Widget  # noqa: F401
from ._version import __version__  # noqa: F401
from .nbextension import _jupyter_nbextension_paths  # noqa: F401

from .graph import Graph  # noqa: F401
from .edge import Edge  # noqa: F401
from .node import Node  # noqa: F401



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from enum import Enum


class Aggregate(Enum):
    '''The aggregation operators available in Perspective. Pass these into the
    `aggregates` arg in `PerspectiveWidget` or `PerspectiveViewer`.

    Examples:
        >>> widget = PerspectiveWidget(data, aggregates={"a": Aggregate.LAST})
    '''
    AND = 'and'
    ANY = 'any'
    AVG = 'avg'
    COUNT = 'count'
    DISTINCT_COUNT = 'distinct count'
    DISTINCT_LEAF = 'distinct leaf'
    DOMINANT = 'dominant'
    FIRST_BY_INDEX = 'first by index'
    LAST_BY_INDEX = 'last by index'
    LAST = 'last'
    HIGH = 'high'
    LOW = 'low'
    MEAN = 'mean'
    MEDIAN = 'median'
    OR = 'or'
    PCT_SUM_PARENT = 'pct sum parent'
    PCT_SUM_GRAND_TOTAL = 'pct sum grand total'
    SUM = 'sum'
    SUM_ABS = 'sum abs'
    SUM_NOT_NULL = 'sum not null'
    UNIQUE = 'unique'

    @staticmethod
    def options():
        return list(c.value for c in Aggregate)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import six
from random import random
from .validate import validate_plugin, validate_columns, validate_row_pivots, validate_column_pivots, \
    validate_aggregates, validate_sort, validate_filters, validate_plugin_config
from .viewer_traitlets import PerspectiveTraitlets
from ..libpsp import is_libpsp

if is_libpsp():
    from ..libpsp import Table, PerspectiveManager


class PerspectiveViewer(PerspectiveTraitlets, object):
    '''PerspectiveViewer wraps the `perspective.Table` API and exposes an API
    around creating views, loading data, and updating data.
    '''

    # Keep track of attributes that can be set via Enum, and their
    # validation methods.
    ENUM_VALIDATORS = {
        "plugin": validate_plugin,
        "aggregates": validate_aggregates,
        "sort": validate_sort
    }

    def __init__(self,
                 plugin='hypergrid',
                 columns=None,
                 row_pivots=None,
                 column_pivots=None,
                 aggregates=None,
                 sort=None,
                 filters=None,
                 plugin_config=None,
                 dark=None,
                 editable=False,
                 manager=None):
        '''Initialize an instance of `PerspectiveViewer` with the given viewer
        configuration.  Do not pass a `Table` or data into the constructor -
        use the :func:`load()` method to provide the viewer with data.

        Keyword Arguments:
            columns (:obj:`list` of :obj:`str`): A list of column names to be
                visible to the user.
            row_pivots (:obj:`list` of :obj:`str`): A list of column names to
                use as row pivots.
            column_pivots (:obj:`list` of :obj:`str`): A list of column names
                to use as column pivots.
            aggregates (:obj:`dict` of :obj:`str` to :obj:`str`):  A dictionary
                of column names to aggregate types, which specify aggregates
                for individual columns.
            sort (:obj:`list` of :obj:`list` of :obj:`str`): A list of lists,
                each list containing a column name and a sort direction
                (``asc``, ``desc``, ``asc abs``, ``desc abs``, ``col asc``,
                ``col desc``, ``col asc abs``, ``col desc abs``).
            filter (:obj:`list` of :obj:`list` of :obj:`str`):  A list of lists,
                each list containing a column name, a filter comparator, and a
                value to filter by.
            plugin (`str`/`perspective.Plugin`): Which plugin to select by
                default.
            plugin_config (`dict`): Custom config for all plugins by name.
            editable (`bool`): Whether to allow editing of the frontend grid.
            manager (`PerspectiveManager`): Manager to use when integrating into more complex systems.

        Examples:
            >>> viewer = PerspectiveViewer(
            ...     aggregates={"a": "avg"},
            ...     row_pivots=["a"],
            ...     sort=[["b", "desc"]],
            ...     filter=[["a", ">", 1]]
            ... )
        '''

        # Create an instance of `PerspectiveManager`, which receives messages
        # from the `PerspectiveJupyterClient` on the front-end.
        if is_libpsp():
            self.manager = manager or PerspectiveManager()
        self.table_name = None  # not a traitlet - only used in python
        self.view_name = None

        # Viewer configuration
        self.plugin = validate_plugin(plugin)
        self.columns = validate_columns(columns) or []
        self.row_pivots = validate_row_pivots(row_pivots) or []
        self.column_pivots = validate_column_pivots(column_pivots) or []
        self.aggregates = validate_aggregates(aggregates) or {}
        self.sort = validate_sort(sort) or []
        self.filters = validate_filters(filters) or []
        self.plugin_config = validate_plugin_config(plugin_config) or {}
        self.dark = dark
        self.editable = editable

    @property
    def table(self):
        '''Returns the `perspective.Table` under management by the viewer.'''
        return self.manager.get_table(self.table_name)

    @property
    def view(self):
        '''Returns the `perspective.View` currently shown by the viewer.

        This property changes every time the viewer configuration changes.'''
        return self.manager.get_view(self.view_name)

    def load(self, table_or_data, **options):
        '''Given a `perspective.Table` or data that can be handled by
        `perspective.Table`, pass it to the viewer.

        `load()` resets the state of the viewer.  If a `perspective.Table`
        is passed into `table_or_data`, `**options` is ignored as the options
        already set on the `Table` take precedence.  If data is passed in, a
        `perspective.Table` is automatically created by this function, and the
        options passed to `**config` are extended to the new Table.

        Args:
            table_or_data (`Table`|`dict`|`list`|`pandas.DataFrame`): a
                `perspective.Table` instance or a dataset to be displayed
                in the viewer.

        Keyword Arguments:
            name (`str`): An optional name to reference the table by so it can
                be accessed from the front-end. If not provided, a name will
                be generated.
            index (`str`): A column name to be used as the primary key.
                Ignored if a `Table` is supplied.
            limit (`int`): A upper limit on the number of rows in the Table.
                Cannot be set at the same time as `index`, ignored if a `Table`
                is passed in.

        Examples:
            >>> from perspective import Table, PerspectiveViewer
            >>> data = {"a": [1, 2, 3]}
            >>> tbl = Table(data)
            >>> viewer = PerspectiveViewer()
            >>> viewer.load(tbl)
            >>> viewer.load(data, index="a") # viewer state is reset
        '''
        name = options.pop("name", str(random()))
        if isinstance(table_or_data, Table):
            table = table_or_data
        else:
            table = Table(table_or_data, **options)

        self.manager.host_table(name, table)

        # Reset the viewer when `load()` is called again.
        if self.table_name is not None:
            self.reset()

        # If the user does not set columns to show, synchronize viewer state
        # with dataset.
        if len(self.columns) == 0:
            self.columns = table.columns()

        self.table_name = name

    def update(self, data):
        '''Update the table under management by the viewer with new data.
        This function follows the semantics of `Table.update()`, and will be
        affected by whether an index is set on the underlying table.

        Args:
            data (`dict`|`list`|`pandas.DataFrame`): the update data for the
                table.
        '''
        self.table.update(data)

    def clear(self):
        '''Clears the rows of this viewer's `Table`.'''
        if self.table is not None:
            self.table.clear()

    def replace(self, data):
        '''Replaces the rows of this viewer's `Table` with new data.

        Args:
            data (`dict`|`list`|`pandas.DataFrame`): new data to set into the
                table - must conform to the table's schema.
        '''
        if self.table is not None:
            self.table.replace(data)

    def save(self):
        '''Get the viewer's attribute as a dictionary, symmetric with `restore` so that a
           viewer's configuration can be reproduced.'''
        return {
            'row_pivots': self.row_pivots,
            'column_pivots': self.column_pivots,
            'filters': self.filters,
            'sort': self.sort,
            'aggregates': self.aggregates,
            'columns': self.columns,
            'plugin': self.plugin,
        }

    def restore(self, **kwargs):
        '''Restore a given set of attributes, passed as kwargs (e.g. dictionary). Symmetric with `save` so that
        a given viewer's configuration can be reproduced'''
        for k, v in six.iteritems(kwargs):
            if k in ('row_pivots', 'column_pivots', 'filters', 'sort', 'aggregates', 'columns', 'plugin'):
                setattr(self, k, v)

    def reset(self):
        '''Resets the viewer's attributes and state, but does not delete or
        modify the underlying `Table`.
        '''
        self.row_pivots = []
        self.column_pivots = []
        self.filters = []
        self.sort = []
        self.aggregates = {}
        self.columns = []
        self.plugin = "hypergrid"

    def delete(self, delete_table=True):
        '''Delete the Viewer's data and clears its internal state. If
        `delete_table` is True, the underlying `perspective.Table` and all
        associated `View`s will be deleted.

        Args:
            delete_table (bool) : whether the underlying `Table` will be
                deleted. Defaults to True.
        '''
        if self.view:
            self.view.delete()
            self.view_name = None

        for view in self.manager._views.values():
            view.delete()

        if delete_table:
            self.table.delete()
            self.manager._tables.pop(self.table_name)
            self.table_name = None

        self.reset()

    def _new_view(self):
        '''Create a new View, and assign its name to the viewer.  Do not call
        this function - it will be called automatically when the state of the
        viewer changes.  There should only be one View associated with the
        Viewer at any given time - when a new View is created, the old one
        is destroyed.
        '''
        if not self.table_name:
            return

        name = str(random())
        table = self.manager.get_table(self.table_name)
        view = table.view(
            row_pivots=self.row_pivots,
            column_pivots=self.column_pivots,
            columns=self.columns,
            aggregates=self.aggregates,
            sort=self.sort,
            filter=self.filters
        )

        self.manager.host_view(name, view)
        self.view_name = name

    def __setattr__(self, name, value):
        """Override __setattr__ in order to allow Enums to be validated
        through Traitlets."""
        if name in PerspectiveViewer.ENUM_VALIDATORS:
            # call the validator and set
            validated = PerspectiveViewer.ENUM_VALIDATORS[name](value)
            value = validated
        super(PerspectiveViewer, self).__setattr__(name, value)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import logging
import json
import random
import string
import datetime
from functools import partial
from ipywidgets import CallbackDispatcher
from ..core.exception import PerspectiveError
from ..table._callback_cache import _PerspectiveCallBackCache
from ..table._date_validator import _PerspectiveDateValidator
from ..table import Table, PerspectiveCppError
from ..table.view import View
from .session import PerspectiveSession

_date_validator = _PerspectiveDateValidator()


class DateTimeEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return _date_validator.to_timestamp(obj)
        else:
            return super(DateTimeEncoder, self).default(obj)


def gen_name(size=10, chars=string.ascii_uppercase + string.digits):
    return "".join(random.choice(chars) for x in range(size))


class PerspectiveManager(object):
    '''PerspectiveManager is an orchestrator for running Perspective on the
    server side.

    The core functionality resides in `process()`, which receives
    JSON-serialized messages from a client (usually `perspective-viewer` in the
    browser), executes the commands in the message, and returns the results of
    those commands back to the `post_callback`.  The manager cannot create
    tables or views - use `host_table` or `host_view` to pass Table/View
    instances to the manager.  Because Perspective is designed to be used in a
    shared context, i.e. multiple clients all accessing the same `Table`,
    PerspectiveManager comes with the context of `sessions` - an
    encapsulation of the actions and resources used by a single connection
    to Perspective.

    - When a client connects, for example through a websocket, a new session
        should be spawned using `new_session()`.
    - When the websocket closes, call `close()` on the session instance to
        clean up associated resources.
    '''

    # Commands that should be blocked from execution when the manager is in
    # `locked` mode, i.e. its tables and views made immutable from remote
    # modification.
    LOCKED_COMMANDS = ["table", "update", "remove", "replace", "clear"]

    def __init__(self, lock=False):
        self._tables = {}
        self._views = {}
        self._callback_cache = _PerspectiveCallBackCache()
        self._queue_process_callback = None
        self._lock = lock

        # ipywidgets-style callbacks (e.g. external callbacks)
        self._update_handlers = CallbackDispatcher()

    def lock(self):
        """Block messages that can mutate the state of `Table`s and `View`s
        under management.

        All `PerspectiveManager`s exposed over the internet should be locked to
        prevent content from being mutated by clients.
        """
        self._lock = True

    def unlock(self):
        """Unblock messages that can mutate the state of `Table`s and `View`s
        under management."""
        self._lock = False

        # ipywidgets-style callbacks (e.g. external callbacks)
        self._update_handlers = CallbackDispatcher()

    def host(self, item, name=None):
        """Given a :obj:`~perspective.Table` or :obj:`~perspective.View`,
        place it under management and allow operations on it to be passed
        through the Manager instance.

        Args:
            table_or_view (:obj:`~perspective.Table`/:obj:`~perspective.View`) :
                a Table or View to be managed.

        Keyword Args:
            name (:obj:`str`) : an optional name to allow retrieval through
                `get_table` or `get_view`. A name will be generated if not
                provided.
        """
        name = name or gen_name()
        if isinstance(item, Table):
            self.host_table(name, item)
        elif isinstance(item, View):
            self.host_view(name, item)
        else:
            raise PerspectiveError(
                "Only `Table()` and `View()` instances can be hosted.")

    def host_table(self, name, table):
        '''Given a reference to a `Table`, manage it and allow operations on it
        to occur through the Manager.

        If a function for `queue_process` is defined (i.e., by
        :obj:`~perspective.PerspectiveTornadoHandler`), bind the function to
        `Table` and have it call the manager's version of `queue_process`.
        '''
        if self._queue_process_callback is not None:
            # always bind the callback to the table's state manager
            table._state_manager.queue_process = partial(
                self._queue_process_callback, state_manager=table._state_manager)
        self._tables[name] = table
        return name

    def host_view(self, name, view):
        '''Given a :obj:`~perspective.View`, add it to the manager's views
        container.
        '''
        self._views[name] = view

    def get_table(self, name):
        '''Return a table under management by name.'''
        return self._tables.get(name, None)

    def get_view(self, name):
        '''Return a view under management by name.'''
        return self._views.get(name, None)

    def new_session(self):
        return PerspectiveSession(self)

    def _set_queue_process(self, func):
        """For each table under management, bind `func` to the table's state
        manager and to run whenever `queue_process` is called.

        After this method is called, future Tables hosted on this manager
        instance will call the same `queue_process` callback.
        """
        self._queue_process_callback = func
        for table in self._tables.values():
            table._state_manager.queue_process = partial(
                self._queue_process_callback, state_manager=table._state_manager)

    def _process(self, msg, post_callback, client_id=None):
        '''Given a message from the client, process it through the Perspective
        engine.

        Args:
            msg (:obj`dict`): a message from the client with instructions
                that map to engine operations
            post_callback (:obj`callable`): a function that returns data to the
                client. `post_callback` should be a callable that takes two
                parameters: `data` (str), and `binary` (bool), a kwarg that
                specifies whether `data` is a binary string.
        '''
        if isinstance(msg, str):
            if msg == "heartbeat":   # TODO fix this
                return
            msg = json.loads(msg)

        if not isinstance(msg, dict):
            raise PerspectiveError(
                "Message passed into `_process` should either be a "
                "JSON-serialized string or a dict.")

        cmd = msg["cmd"]

        if self._is_locked_command(msg) is True:
            error_message = "`{0}` failed - access denied".format(
                msg["cmd"] + (("." + msg["method"]) if msg.get("method", None) is not None else ""))
            post_callback(json.dumps(self._make_error_message(
                msg["id"], error_message), cls=DateTimeEncoder))
            return

        try:
            if cmd == "init":
                # return empty response
                post_callback(
                    json.dumps(
                        self._make_message(
                            msg["id"],
                            None),
                        cls=DateTimeEncoder))
            elif cmd == "table":
                try:
                    # create a new Table and track it
                    data_or_schema = msg["args"][0]
                    self._tables[msg["name"]] = Table(
                        data_or_schema, **msg.get("options", {}))
                except IndexError:
                    self._tables[msg["name"]] = []
            elif cmd == "view":
                # create a new view and track it with the assigned client_id.
                new_view = self._tables[msg["table_name"]].view(
                    **msg.get("config", {}))
                new_view._client_id = client_id
                self._views[msg["view_name"]] = new_view
            elif cmd == "table_method" or cmd == "view_method":
                self._process_method_call(msg, post_callback, client_id)
        except(PerspectiveError, PerspectiveCppError) as e:
            # Catch errors and return them to client
            post_callback(
                json.dumps(
                    self._make_error_message(
                        msg["id"],
                        str(e))),
                cls=DateTimeEncoder)

    def _process_method_call(self, msg, post_callback, client_id):
        '''When the client calls a method, validate the instance it calls on
        and return the result.
        '''
        if msg["cmd"] == "table_method":
            table_or_view = self._tables.get(msg["name"], None)
        else:
            table_or_view = self._views.get(msg["name"], None)
            if table_or_view is None:
                post_callback(
                    json.dumps(
                        self._make_error_message(
                            msg["id"],
                            "View is not initialized"),
                        cls=DateTimeEncoder))
        try:
            if msg.get("subscribe", False) is True:
                self._process_subscribe(
                    msg, table_or_view, post_callback, client_id)
            else:
                args = {}
                if msg["method"] == "schema":
                    # make sure schema returns string types
                    args["as_string"] = True
                elif msg["method"].startswith("to_"):
                    # parse options in `to_format` calls
                    for d in msg.get("args", []):
                        args.update(d)
                else:
                    args = msg.get("args", [])

                if msg["method"] == "delete" and msg["cmd"] == "view_method":
                    # views can be removed, but tables cannot
                    self._views[msg["name"]].delete()
                    self._views.pop(msg["name"], None)
                    return

                # defer to callbacks
                if msg["cmd"] == "table_method" and msg["method"] == "update":
                    self._update_handlers(self, msg)

                if msg["method"].startswith("to_"):
                    # to_format takes dictionary of options
                    result = getattr(table_or_view, msg["method"])(**args)
                elif msg["method"] != "delete":
                    # otherwise parse args as list
                    result = getattr(table_or_view, msg["method"])(*args)

                if isinstance(result, bytes) and msg["method"] != "to_csv":
                    # return a binary to the client without JSON serialization,
                    # i.e. when we return an Arrow. If a method is added that
                    # returns a string, this condition needs to be updated as
                    # an Arrow binary is both `str` and `bytes` in Python 2.
                    self._process_bytes(result, msg, post_callback)
                else:
                    # return the result to the client
                    post_callback(json.dumps(self._make_message(
                            msg["id"], result), cls=DateTimeEncoder))
        except Exception as error:
            post_callback(json.dumps(self._make_error_message(
                msg["id"], str(error)), cls=DateTimeEncoder))

    def _process_subscribe(self, msg, table_or_view, post_callback, client_id):
        '''When the client attempts to add or remove a subscription callback,
        validate and perform the requested operation.

        Args:
            msg (dict): the message from the client
            table_or_view {Table|View} : the instance that the subscription
                will be called on.
            post_callback (callable): a method that notifies the client with
                new data.
            client_id (str) : a unique str id that identifies the
                `PerspectiveSession` object that is passing the message.
        '''
        try:
            callback = None
            callback_id = msg.get("callback_id", None)
            method = msg.get("method", None)
            args = msg.get("args", [])
            if method and method[:2] == "on":
                # wrap the callback
                callback = partial(
                    self.callback, msg=msg, post_callback=post_callback)
                if callback_id:
                    self._callback_cache.add_callback({
                        "client_id": client_id,
                        "callback_id": callback_id,
                        "callback": callback,
                        "name": msg.get("name", None)
                    })
            elif callback_id is not None:
                # remove the callback with `callback_id`
                self._callback_cache.remove_callbacks(
                    lambda cb: cb["callback_id"] != callback_id)
            if callback is not None:
                # call the underlying method on the Table or View, and apply
                # the dictionary of kwargs that is passed through.
                if (method == "on_update"):
                    # If there are no arguments, make sure we call `on_update`
                    # with mode set to "none".
                    mode = {"mode": "none"}
                    if len(args) > 0:
                        mode = args[0]
                    getattr(table_or_view, method)(callback, **mode)
                else:
                    getattr(table_or_view, method)(callback)
            else:
                logging.info("callback not found for remote call {}".format(msg))
        except Exception as error:
            post_callback(json.dumps(self._make_error_message(msg["id"], error), cls=DateTimeEncoder))

    def _process_bytes(self, binary, msg, post_callback):
        """Send a bytestring message to the client without attempting to
        serialize as JSON.

        Perspective's client expects two messages to be sent when a binary
        payload is expected: the first message is a JSON-serialized string with
        the message's `id` and `msg`, and the second message is a bytestring
        without any additional metadata. Implementations of the `post_callback`
        should have an optional kwarg named `binary`, which specifies whether
        `data` is a bytestring or not.

        Args:
            binary (bytes, bytearray) : a byte message to be passed to the client.
            msg (dict) : the original message that generated the binary
                response from Perspective.
            post_callback (callable) : a function that passes data to the
                client, with a `binary` (bool) kwarg that allows it to pass
                byte messages without serializing to JSON.
        """
        pre_msg = self._make_message(msg["id"], "")
        pre_msg["is_transferable"] = True
        post_callback(json.dumps(pre_msg, cls=DateTimeEncoder))
        post_callback(binary, binary=True)

    def callback(self, *args, **kwargs):
        '''Return a message to the client using the `post_callback` method.'''
        id = kwargs.get("msg")["id"]
        data = kwargs.get("event", None)
        post_callback = kwargs.get("post_callback")
        msg = self._make_message(id, data)
        if len(args) > 0 and type(args[0]) == bytes:
            self._process_bytes(args[0], msg, post_callback)
        else:
            post_callback(json.dumps(msg, cls=DateTimeEncoder))

    def clear_views(self, client_id):
        '''Garbage collect views that belong to closed connections.'''
        count = 0
        names = []

        if not client_id:
            raise PerspectiveError("Cannot garbage collect views that are not linked to a specific client ID!")

        for name, view in self._views.items():
            if view._client_id == client_id:
                view.delete()
                names.append(name)
                count += 1

        for name in names:
            self._views.pop(name)

        logging.warning("GC {} views in memory".format(count))

    def _make_message(self, id, result):
        '''Return a serializable message for a successful result.'''
        return {
            "id": id,
            "data": result
        }

    def _make_error_message(self, id, error):
        '''Return a serializable message for an error result.'''
        return {
            "id": id,
            "error": error
        }

    def on_update(self, callback, remove=False):
        '''register an update callback from the frontend'''
        self._update_handlers.register_callback(callback, remove=remove)

    def _is_locked_command(self, msg):
        '''Returns `True` if the manager instance is locked and the command
        is in `PerspectiveManager.LOCKED_COMMANDS`, and `False` otherwise.'''
        if not self._lock:
            return False

        cmd = msg["cmd"]
        method = msg.get("method", None)

        if cmd == "table_method" and method == "delete":
            # table.delete is blocked, but view.delete is not
            return True

        return cmd == "table" or method in PerspectiveManager.LOCKED_COMMANDS




def split(data, callback1, callback2):
    '''Pass data to 2 callbacks

    Args:
        data (any): data to pass to both callbacks
        callback1 (callable): first function to call
        callback2 (callable): second function to call
    '''
    return (callback1(data), callback2(data))


def map(data, *callbacks):
    '''Pass data to multiple callbacks

    Args:
        data (any): data to pass to all callbacks
        callbacks (tuple): callbacks to pass data to
    '''
    return (callback(data) for callback in callbacks)


def merge(data1, data2, callback):
    '''merge two data sources into one callback

    Args:
        data1 (any): first data to pass to callback
        data2 (any): second data to pass to callback
        callback (callable): callback to pass data to
    '''
    return callback((data1, data2))


def reduce(callback, *datas):
    '''merge multiple data sources into one callback

    Args:
        callback (callable): callback to pass data to
        datas (tuple): data to pass to callback
    '''
    return callback((data for data in datas))



import tributary.lazy as tl


class TestAPI:
    def test_api(self):
        assert hasattr(tl.LazyNode, '__add__')
        assert hasattr(tl.LazyNode, '__add__')
        assert hasattr(tl.LazyNode, '__radd__')
        assert hasattr(tl.LazyNode, '__sub__')
        assert hasattr(tl.LazyNode, '__rsub__')
        assert hasattr(tl.LazyNode, '__mul__')
        assert hasattr(tl.LazyNode, '__rmul__')
        assert hasattr(tl.LazyNode, '__div__')
        assert hasattr(tl.LazyNode, '__rdiv__')
        assert hasattr(tl.LazyNode, '__truediv__')
        assert hasattr(tl.LazyNode, '__rtruediv__')
        assert hasattr(tl.LazyNode, '__pow__')
        assert hasattr(tl.LazyNode, '__rpow__')
        assert hasattr(tl.LazyNode, '__mod__')
        assert hasattr(tl.LazyNode, '__rmod__')
        assert hasattr(tl.LazyNode, 'sum')
        assert hasattr(tl.LazyNode, 'average')
        assert hasattr(tl.LazyNode, 'invert')
        assert hasattr(tl.LazyNode, '__and__')
        assert hasattr(tl.LazyNode, '__or__')
        assert hasattr(tl.LazyNode, '__invert__')
        assert hasattr(tl.LazyNode, '__bool__')
        assert hasattr(tl.LazyNode, 'int')
        assert hasattr(tl.LazyNode, 'float')
        assert hasattr(tl.LazyNode, '__str__')
        assert hasattr(tl.LazyNode, '__lt__')
        assert hasattr(tl.LazyNode, '__le__')
        assert hasattr(tl.LazyNode, '__gt__')
        assert hasattr(tl.LazyNode, '__ge__')
        assert hasattr(tl.LazyNode, '__eq__')
        assert hasattr(tl.LazyNode, '__ne__')
        assert hasattr(tl.LazyNode, '__neg__')
        assert hasattr(tl.LazyNode, '__len__')
        assert hasattr(tl.LazyNode, 'log')
        assert hasattr(tl.LazyNode, 'sin')
        assert hasattr(tl.LazyNode, 'cos')
        assert hasattr(tl.LazyNode, 'tan')
        assert hasattr(tl.LazyNode, 'asin')
        assert hasattr(tl.LazyNode, 'acos')
        assert hasattr(tl.LazyNode, 'atan')
        assert hasattr(tl.LazyNode, 'abs')
        assert hasattr(tl.LazyNode, 'sqrt')
        assert hasattr(tl.LazyNode, 'exp')
        assert hasattr(tl.LazyNode, 'erf')
        assert hasattr(tl.LazyNode, 'round')
        assert hasattr(tl.LazyNode, 'floor')
        assert hasattr(tl.LazyNode, 'ceil')

        # assert hasattr(tl.LazyNode, 'delay')
        # assert hasattr(tl.LazyNode, 'state')
        # assert hasattr(tl.LazyNode, 'apply')
        # assert hasattr(tl.LazyNode, 'window')
        # assert hasattr(tl.LazyNode, 'unroll')
        # assert hasattr(tl.LazyNode, 'unrollDataFrame')
        # assert hasattr(tl.LazyNode, 'merge')
        # assert hasattr(tl.LazyNode, 'listMerge')
        # assert hasattr(tl.LazyNode, 'dictMerge')
        # assert hasattr(tl.LazyNode, 'reduce')

        assert hasattr(tl.LazyNode, 'graph')
        # assert hasattr(tl.LazyNode, 'pprint')
        assert hasattr(tl.LazyNode, 'graphviz')
        assert hasattr(tl.LazyNode, 'dagre')
        assert hasattr(tl.LazyNode, 'print')

    def test_api_lazy_specific(self):
        '''apis specific to lazy nodes'''
        assert hasattr(tl.LazyNode, 'expire')
        assert hasattr(tl.LazyNode, 'interval')



import math
import numpy as np
import scipy as sp
from .utils import _CALCULATIONS_GRAPHVIZSHAPE
from ..base import Node, _gen_node


def unary(foo, name):
    def _foo(self):
        downstream = Node(foo, {}, name=name, inputs=1, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
        self._downstream.append((downstream, 0))
        downstream._upstream.append(self)
        return downstream
    return _foo


def binary(foo, name):
    def _foo(self, other):
        other = _gen_node(other)
        downstream = Node(foo, {}, name=name, inputs=2, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
        self._downstream.append((downstream, 0))
        other._downstream.append((downstream, 1))
        downstream._upstream.extend([self, other])
        return downstream
    return _foo


def n_ary(foo, name):
    def _foo(*others):
        downstream = Node(foo, {}, name=name, inputs=len(others), graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
        for i, other in enumerate(others):
            other._downstream.append((downstream, i))

        downstream._upstream.extend(list(others))
        return downstream
    return _foo


########################
# Arithmetic Operators #
########################
Noop = unary(lambda x: x, name='Noop')
Negate = unary(lambda x: -1 * x, name='Negate')
Invert = unary(lambda x: 1 / x, name='Invert')
Add = binary(lambda x, y: x + y, name='Add')
Sub = binary(lambda x, y: x - y, name='Sub')
Mult = binary(lambda x, y: x * y, name='Mult')
Div = binary(lambda x, y: x / y, name='Div')
RDiv = binary(lambda x, y: y / x, name='RDiv')
Mod = binary(lambda x, y: x % y, name='Mod')
Pow = binary(lambda x, y: x ** y, name='Pow')
Sum = n_ary(lambda *args: sum(args), name='Sum')
Average = n_ary(lambda *args: sum(args) / len(args), name='Average')

#####################
# Logical Operators #
#####################
Not = unary(lambda x: not x, name='Not')
And = binary(lambda x, y: x and y, name='And')
Or = binary(lambda x, y: x or y, name='Or')


###################
# Numpy Functions #
###################
def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
    if ufunc == np.add:
        if isinstance(inputs[0], Node):
            return inputs[0].__add__(inputs[1])
        else:
            return inputs[1].__add__(inputs[0])
    elif ufunc == np.subtract:
        if isinstance(inputs[0], Node):
            return inputs[0].__sub__(inputs[1])
        else:
            return inputs[1].__sub__(inputs[0])
    elif ufunc == np.multiply:
        if isinstance(inputs[0], Node):
            return inputs[0].__mul__(inputs[1])
        else:
            return inputs[1].__mul__(inputs[0])
    elif ufunc == np.divide:
        if isinstance(inputs[0], Node):
            return inputs[0].__truediv__(inputs[1])
        else:
            return inputs[1].__truediv__(inputs[0])
    elif ufunc == np.sin:
        return inputs[0].sin()
    elif ufunc == np.cos:
        return inputs[0].cos()
    elif ufunc == np.tan:
        return inputs[0].tan()
    elif ufunc == np.arcsin:
        return inputs[0].asin()
    elif ufunc == np.arccos:
        return inputs[0].acos()
    elif ufunc == np.arctan:
        return inputs[0].atan()
    elif ufunc == np.exp:
        return inputs[0].exp()
    elif ufunc == sp.special.erf:
        return inputs[0].erf()
    raise NotImplementedError('Not Implemented!')


def __array_function__(self, func, method, *inputs, **kwargs):
    if func == np.lib.scimath.log:
        return inputs[0][0].log()
    elif func == np.lib.scimath.sqrt:
        return inputs[0][0].sqrt()
    raise NotImplementedError('Not Implemented!')


###############
# Comparators #
###############
Equal = binary(lambda x, y: x == y, name='Equal')
NotEqual = binary(lambda x, y: x != y, name='NotEqual')
Lt = binary(lambda x, y: x < y, name='Less')
Le = binary(lambda x, y: x <= y, name='LessOrEqual')
Gt = binary(lambda x, y: x > y, name='Greater')
Ge = binary(lambda x, y: x >= y, name='GreaterOrEqual')


##########################
# Mathematical Functions #
##########################
Log = unary(lambda x: math.log(x), name='Log')
Sin = unary(lambda x: math.sin(x), name='Sin')
Cos = unary(lambda x: math.cos(x), name='Cos')
Tan = unary(lambda x: math.tan(x), name='Tan')
Arcsin = unary(lambda x: math.asin(x), name='Arcsin')
Arccos = unary(lambda x: math.acos(x), name='Arccos')
Arctan = unary(lambda x: math.atan(x), name='Arctan')
Sqrt = unary(lambda x: math.sqrt(x), name='Sqrt')
Abs = unary(lambda x: abs(x), name='Abs')
Exp = unary(lambda x: math.exp(x), name='Exp')
Erf = unary(lambda x: math.erf(x), name='Erf')
Floor = unary(lambda x: math.floor(x), name='Floor')
Ceil = unary(lambda x: math.ceil(x), name='Ceil')


def Round(self, ndigits=0):
    downstream = Node(lambda x: round(x, ndigits=ndigits), {}, name="Round", inputs=1, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
    self._downstream.append((downstream, 0))
    downstream._upstream.append(self)
    return downstream


##############
# Converters #
##############
Int = unary(lambda x: int(x), name='Int')
Float = unary(lambda x: float(x), name='Float')
Bool = unary(lambda x: bool(x), name='Bool')
Str = unary(lambda x: str(x), name='Str')


# __Bool__ = unary(lambda x: bool(x), name='Noop')

def __Bool__(self):
    return True


###################
# Python Builtins #
###################
Len = unary(lambda x: len(x), name='Len')


########################
# Arithmetic Operators #
########################
Node.__add__ = Add
Node.__radd__ = Add
Node.__sub__ = Sub
Node.__rsub__ = Sub
Node.__mul__ = Mult
Node.__rmul__ = Mult
# Node.__matmul__ = MatMult
# Node.__rmatmul__ = MatMult
Node.__div__ = Div
Node.__rdiv__ = RDiv
# Node.__divmod__ = DivMod
# Node.__rdivmod__ = DivMod
Node.__truediv__ = Div
Node.__rtruediv__ = RDiv
Node.__floordiv__ = Div
# Node.__lshift__ = LShift
# Node.__rlshift__ = LShift
# Node.__rshift__ = RShift
# Node.__rrshift__ = RShift

Node.__pow__ = Pow
Node.__rpow__ = Pow
Node.__mod__ = Mod
Node.__rmod__ = Mod

Node.sum = Sum
Node.average = Average
Node.invert = Invert


#####################
# Logical Operators #
#####################
Node.__and__ = And
Node.__rand__ = And
Node.__or__ = Or
Node.__ror__ = Or
# Node.__xor__ = Xor
# Node.__rxor__ = Xor
Node.__invert__ = Not
Node.__bool__ = __Bool__

##############
# Converters #
##############
Node.int = Int
# Node.__int__ = Int
Node.float = Float
# Node.__float__ = Float
# Node.__str__ = Str
# Node.__complex__ = Int

###############
# Comparators #
###############
Node.__lt__ = Lt
Node.__le__ = Le
Node.__gt__ = Gt
Node.__ge__ = Ge
Node.__eq__ = Equal
Node.__ne__ = NotEqual
Node.__neg__ = Negate
# Node.__pos__ =
# Node.__nonzero__ = Bool  # Py2 compat

###################
# Python Builtins #
###################
Node.__len__ = Len
# Node.__round__ = Len
# Node.__trunc__ = Len
# Node.__floor__ = Len
# Node.__ceil__ = Len

###################
# Numpy Functions #
###################
Node.__array_ufunc__ = __array_ufunc__
Node.__array_function__ = __array_function__


##########################
# Mathematical Functions #
##########################
Node.log = Log
Node.sin = Sin
Node.cos = Cos
Node.tan = Tan
Node.asin = Arcsin
Node.acos = Arccos
Node.atan = Arctan
Node.abs = Abs
Node.sqrt = Sqrt
Node.exp = Exp
Node.erf = Erf
Node.floor = Floor
Node.ceil = Ceil
Node.round = Round



# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from perspective.table import Table


class TestRemove(object):

    def test_remove_all(self):
        tbl = Table([{"a": "abc", "b": 123}], index="a")
        tbl.remove(["abc"])
        assert tbl.view().to_records() == []

    def test_remove_nonsequential(self):
        tbl = Table([{"a": "abc", "b": 123}, {"a": "def", "b": 456}, {"a": "efg", "b": 789}], index="a")
        tbl.remove(["abc", "efg"])
        assert tbl.view().to_records() == [{"a": "def", "b": 456}]

    def test_remove_multiple_single(self):
        tbl = Table({"a": int, "b": str}, index="a")
        for i in range(0, 10):
            tbl.update([{"a": i, "b": str(i)}])
        for i in range(1, 10):
            tbl.remove([i])
        assert tbl.view().to_records() == [{"a": 0, "b": "0"}]



import aiofiles
import json as JSON
from .output import _OUTPUT_GRAPHVIZSHAPE
from ..base import Node


def File(node, filename='', json=True):
    '''Open up a file and write lines to the file

    Args:
        node (Node): input stream
        filename (str): filename to write
        json (bool): load file line as json
    '''

    async def _file(data):
        async with aiofiles.open(filename, mode='a') as fp:
            if json:
                fp.write(JSON.dumps(data))
            else:
                fp.write(data)
        return data

    ret = Node(foo=_file, name='File', inputs=1, graphvizshape=_OUTPUT_GRAPHVIZSHAPE)
    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret



from traitlets import HasTraits, observe, Bool, Instance, List, Dict
from .edge import Edge
from .node import Node


class Graph(HasTraits):
    directed = Bool(default_value=True)
    multigraph = Bool(default_value=False)
    compound = Bool(default_value=False)
    attrs = Dict()

    edges = List(trait=Instance(Edge))
    nodes = List(trait=Instance(Node))

    _widget = None

    def __init__(self, directed=True, multigraph=False, compound=False, attrs=None):
        super(Graph, self).__init__(directed=directed, multigraph=multigraph, compound=compound, attrs=attrs or {})

    def isDirected(self):
        return self.directed

    def isMultiGraph(self):
        return self.multigraph

    def setGraph(self, config):
        config = config or {}
        self.directed = config.get("directed", self.directed)
        self.multigraph = config.get("multigraph", self.multigraph)
        self.compound = config.get("compound", self.compound)
        self.attrs = config.get("attrs", self.attrs)

    def edgeCount(self):
        return len(self.edges)

    def hasEdge(self, edge):
        if isinstance(edge, Edge):
            return edge in self.edges
        elif isinstance(edge, tuple):
            if isinstance(edge[0], Node):
                return any((e for e in self.edges if e.v == edge[0] and e.w == edge[1]))
            else:
                return any((e for e in self.edges if e.v.name == edge[0] and e.w.name == edge[1]))
        return edge in (e.label for e in self.edges)

    # inEdges(inNodeName: string, outNodeName?: string): Edge[]|undefined;
    # outEdges(outNodeName: string, inNodeName?: string): Edge[]|undefined;
    # removeEdge(outNodeName: string, inNodeName: string): Graph;
    # setDefaultEdgeLabel(callback: string|((v: string, w: string, name?: string) => string|Label)): Graph;
    def setEdge(self, edge_or_node1, node2=None, label="", **attrs):
        if isinstance(edge_or_node1, Edge):
            if edge_or_node1.v not in self.nodes:
                self.setNode(edge_or_node1.v)
            if edge_or_node1.w not in self.nodes:
                self.setNode(edge_or_node1.w)
            self.edges.append(edge_or_node1)
            edge_or_node1['label'] = label or edge_or_node1.label
            edge_or_node1['tooltip'] = attrs.pop("tooltip", edge_or_node1.tooltip)
            edge_or_node1['labelpos'] = attrs.pop("labelpos", edge_or_node1.labelpos)
            edge_or_node1['labeloffset'] = attrs.pop("labeloffset", edge_or_node1.labeloffset)
            edge_or_node1['attrs'].update(attrs)
            return edge_or_node1
        if isinstance(edge_or_node1, Node):
            node1 = edge_or_node1
        else:
            # create and add node
            node1 = self.setNode(edge_or_node1)
        if not isinstance(node2, Node):
            node2 = self.setNode(node2)

        edge = Edge(node1, node2, label=label, **attrs)
        edge._setGraph(self)
        self.edges.append(edge)
        self.post({"type": "setEdge",
                   "source": edge.to_dict()})
        return edge

    # TODO
    # children(parentName: string): string|undefined;

    def hasNode(self, node):
        if isinstance(node, Node):
            return node in self.nodes
        return node in (n.name for n in self.nodes)

    # neighbors(name: string): Node[]|undefined;

    def nodeCount(self):
        return len(self.nodes)

    # parent(childName: string): string|undefined;
    # predecessors(name: string): Node[]|undefined;
    # removeNode(name: string): Graph;
    # setDefaultNodeLabel(callback: string|((nodeId: string) => string|Label)): Graph;

    def setNode(self, node, **attrs):
        if not isinstance(node, Node):
            # find existing
            for n in self.nodes:
                if n.name == node:
                    node = n
                    break
            # doesnt exist, make new
            if not isinstance(node, Node):
                node = Node(node, **attrs)
                self.nodes.append(node)
            else:
                node.shape = attrs.pop("shape", node.shape)
                node.label = attrs.pop("label", node.label)
                node.tooltip = attrs.pop("tooltip", node.tooltip)
                node.attrs.update(attrs)
        else:
            if node not in self.nodes:
                self.nodes.append(node)
            else:
                node.shape = attrs.pop("shape", node.shape)
                node.label = attrs.pop("label", node.label)
                node.tooltip = attrs.pop("tooltip", node.tooltip)
                node.attrs.update(attrs)

        node._setGraph(self)
        self.post({"type": "setNode",
                   "source": node.to_dict()})
        return node

    # setParent(childName: string, parentName: string): void;
    # sinks(): Node[];
    # sources(): Node[];
    # successors(name: string): Node[]|undefined;

    # internal
    def _setWidget(self, w):
        self._widget = w

    def post(self, msg):
        if self._widget:
            self._widget.post(msg)
        return

    def _notify_change(self, source, attr, value):
        if isinstance(source, Edge):
            self.post({"type": "edge",
                       "source": source.to_dict()})

        elif isinstance(source, Node):
            self.post({"type": "node",
                       "source": source.to_dict()})
        elif source == self:
            self.post({"type": "graph",
                       "source": self.to_dict()})
        else:
            raise Exception('Unknown source: {}'.format(source))

    @observe('directed')
    def _observe_directed(self, change):
        self._notify_change(self, 'directed', change['new'])

    @observe('multigraph')
    def _observe_multigraph(self, change):
        self._notify_change(self, 'multigraph', change['new'])

    @observe('compound')
    def _observe_compound(self, change):
        self._notify_change(self, 'compound', change['new'])

    @observe('attrs')
    def _observe_attrs(self, change):
        self._notify_change(self, 'attrs', change['new'])

    def to_dict(self):
        ret = {}
        ret["directed"] = self.directed
        ret["multigraph"] = self.multigraph
        ret["compound"] = self.compound
        ret["attrs"] = self.attrs
        ret["nodes"] = [n.to_dict() for n in self.nodes]
        ret["edges"] = [e.to_dict() for e in self.edges]
        return ret



# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import numpy as np
import pandas as pd
from datetime import date, datetime
from pytest import mark
from perspective.table import Table


class TestUpdatePandas(object):
    def test_update_df(self):
        tbl = Table({"a": [1, 2, 3, 4]})

        update_data = pd.DataFrame({
            "a": [5, 6, 7, 8]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [1, 2, 3, 4, 5, 6, 7, 8]
        }

    def test_update_df_i32_vs_i64(self):
        tbl = Table({"a": int})

        update_data = pd.DataFrame({
            "a": np.array([5, 6, 7, 8], dtype="int64")
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [5, 6, 7, 8]
        }

    def test_update_df_bool(self):
        tbl = Table({"a": [True, False, True, False]})

        update_data = pd.DataFrame({
            "a": [True, False, True, False]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [True, False, True, False, True, False, True, False]
        }

    def test_update_df_str(self):
        tbl = Table({"a": ["a", "b", "c", "d"]})

        update_data = pd.DataFrame({
            "a": ["a", "b", "c", "d"]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": ["a", "b", "c", "d", "a", "b", "c", "d"]
        }

    def test_update_df_date(self):
        tbl = Table({"a": [date(2019, 7, 11)]})

        update_data = pd.DataFrame({
            "a": [date(2019, 7, 12)]
        })

        tbl.update(update_data)
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 11), datetime(2019, 7, 12)]
        }

    def test_update_df_date_timestamp(self, util):
        tbl = Table({"a": [date(2019, 7, 11)]})

        assert tbl.schema() == {
            "a": date
        }

        update_data = pd.DataFrame({
            "a": [util.to_timestamp(datetime(2019, 7, 12, 0, 0, 0))]
        })

        tbl.update(update_data)
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 11), datetime(2019, 7, 12)]
        }

    def test_update_df_datetime(self):
        tbl = Table({"a": [np.datetime64(datetime(2019, 7, 11, 11, 0))]})

        update_data = pd.DataFrame({
            "a": [datetime(2019, 7, 12, 11, 0)]
        })

        tbl.update(update_data)
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 11, 11, 0), datetime(2019, 7, 12, 11, 0)]
        }

    def test_update_df_datetime_timestamp_seconds(self, util):
        tbl = Table({"a": [np.datetime64(datetime(2019, 7, 11, 11, 0))]})

        update_data = pd.DataFrame({
            "a": [util.to_timestamp(datetime(2019, 7, 12, 11, 0))]
        })

        tbl.update(update_data)
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 11, 11, 0), datetime(2019, 7, 12, 11, 0)]
        }

    def test_update_df_datetime_timestamp_ms(self, util):
        tbl = Table({"a": [np.datetime64(datetime(2019, 7, 11, 11, 0))]})

        update_data = pd.DataFrame({
            "a": [util.to_timestamp(datetime(2019, 7, 12, 11, 0)) * 1000]
        })

        tbl.update(update_data)
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 11, 11, 0), datetime(2019, 7, 12, 11, 0)]
        }

    def test_update_df_partial(self):
        tbl = Table({
            "a": [1, 2, 3, 4],
            "b": ["a", "b", "c", "d"]
        }, index="b")

        update_data = pd.DataFrame({
            "a": [5, 6, 7, 8],
            "b": ["a", "b", "c", "d"]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [5, 6, 7, 8],
            "b": ["a", "b", "c", "d"]
        }

    def test_update_df_partial_implicit(self):
        tbl = Table({"a": [1, 2, 3, 4]})

        update_data = pd.DataFrame({
            "a": [5, 6, 7, 8],
            "__INDEX__": [0, 1, 2, 3]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [5, 6, 7, 8]
        }

    def test_update_df_datetime_partial(self):
        tbl = Table({
            "a": [np.datetime64(datetime(2019, 7, 11, 11, 0))],
            "b": [1]
        }, index="b")

        update_data = pd.DataFrame({
            "a": [datetime(2019, 7, 12, 11, 0)],
            "b": [1]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 7, 12, 11, 0)],
            "b": [1]
        }

    def test_update_df_one_col(self):
        tbl = Table({
            "a": [1, 2, 3, 4],
            "b": ["a", "b", "c", "d"]
        })

        update_data = pd.DataFrame({
            "a": [5, 6, 7]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [1, 2, 3, 4, 5, 6, 7],
            "b": ["a", "b", "c", "d", None, None, None]
        }

    def test_update_df_nonseq_partial(self):
        tbl = Table({
            "a": [1, 2, 3, 4],
            "b": ["a", "b", "c", "d"]
        }, index="b")

        update_data = pd.DataFrame({
            "a": [5, 6, 7],
            "b": ["a", "c", "d"]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [5, 2, 6, 7],
            "b": ["a", "b", "c", "d"]
        }

    def test_update_df_with_none_partial(self):
        tbl = Table({
            "a": [1, np.nan, 3],
            "b": ["a", None, "d"]
        }, index="b")

        update_data = pd.DataFrame({
            "a": [4, 5],
            "b": ["a", "d"]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [None, 4, 5],
            "b": [None, "a", "d"]  # pkeys are ordered
        }

    def test_update_df_unset_partial(self):
        tbl = Table({
            "a": [1, 2, 3],
            "b": ["a", "b", "c"]
        }, index="b")

        update_data = pd.DataFrame({
            "a": [None, None],
            "b": ["a", "c"]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [None, 2, None],
            "b": ["a", "b", "c"]
        }

    def test_update_df_nan_partial(self):
        tbl = Table({
            "a": [1, 2, 3],
            "b": ["a", "b", "c"]
        }, index="b")

        update_data = pd.DataFrame({
            "a": [None, None],
            "b": ["a", "c"]
        })

        tbl.update(update_data)

        assert tbl.view().to_dict() == {
            "a": [None, 2, None],
            "b": ["a", "b", "c"]
        }



import os
import os.path
import tributary.streaming as ts


class TestFile:
    def test_file(self):
        file = os.path.abspath(os.path.join(os.path.dirname(__file__), 'test_file_data.csv'))

        def foo():
            yield 1
            yield 2
            yield 3
            yield 4

        out = ts.FileSink(ts.Foo(foo), filename=file)
        assert ts.run(out) == [1, 2, 3, 4]



from .base import LazyGraph, Node as LazyNode, node  # noqa: F401
from .calculations import *  # noqa: F401, F403
from .input import *  # noqa: F401, F403
from .output import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
from logging import critical


def is_libpsp():
    """Was libbinding successfully loaded in this module?"""
    return __is_libpsp__


__is_libpsp__ = True

try:
    # Load all `libbinding` depending modules in one go, otherwise nothing
    # dependent on `libbinding` is exposed.
    from .table import *  # noqa: F401, F403
    from .manager import *  # noqa: F401, F403
    from .tornado_handler import *  # noqa: F401, F403
    from .viewer import *  # noqa: F401, F403
except ImportError:
    __is_libpsp__ = False
    critical("Failed to import C++ bindings for Perspective "
             "probably as it could not be built for your architecture "
             "(check install logs for more details).\n"
             "You can still use `PerspectiveWidget` in client mode using JupyterLab.", exc_info=True)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import six
import tornado.ioloop
from functools import partial
from threading import Thread
from pytest import mark
from perspective import Table, PerspectiveManager
from perspective.table._state import _PerspectiveStateManager


class AsyncSentinel(object):
    """Check how many times the callback is actually being called
    on the IOLoop."""

    def __init__(self, value):
        self.value = value

    def set(self, new):
        self.value = new

    def get(self):
        return self.value


SENTINEL = AsyncSentinel(0)


def queue_process_async(table_id, state_manager, loop=None):
    """Create our own `queue_process` method that uses a Tornado IOLoop."""
    if loop:
        SENTINEL.set(SENTINEL.get() + 1)
        loop.add_callback(state_manager.call_process, table_id=table_id)


def queue_process_async_delay(table_id, state_manager, delay=0.25, loop=None):
    """Create our own `queue_process` method that uses a Tornado IOLoop."""
    if loop:
        SENTINEL.set(SENTINEL.get() + 1)
        loop.call_later(delay, state_manager.call_process, table_id=table_id)


data = [{"a": i, "b": i * 0.5, "c": str(i)} for i in range(10)]


# @mark.skipif(six.PY2, reason="Requires Python 3")
class TestAsync(object):

    @classmethod
    def setup_class(cls):
        cls.loop = tornado.ioloop.IOLoop()
        cls.loop.make_current()
        cls.wrapped_queue_process = partial(queue_process_async, loop=cls.loop)
        cls.thread = Thread(target=cls.loop.start)
        cls.thread.daemon = True
        cls.thread.start()

    @classmethod
    def teardown_class(cls):
        cls.loop.add_callback(lambda: tornado.ioloop.IOLoop.current().stop())
        cls.loop.clear_current()
        cls.thread.join()
        cls.loop.close(all_fds=True)

    @classmethod
    def loop_is_running(cls):
        if six.PY2:
            return cls.loop._running
        else:
            return cls.loop.asyncio_loop.is_running()

    def setup_method(self):
        global SENTINEL
        SENTINEL = AsyncSentinel(0)

    def teardown_method(self):
        global SENTINEL
        SENTINEL = AsyncSentinel(0)

    def test_async_queue_process(self):
        tbl = Table({
            "a": int,
            "b": float,
            "c": str
        })
        manager = PerspectiveManager()
        manager._set_queue_process(TestAsync.wrapped_queue_process)
        manager.host(tbl)

        assert tbl.size() == 0

        for i in range(5):
            tbl.update([data[i]])

        table_id = tbl._table.get_id()
        pool = tbl._table.get_pool()

        assert _PerspectiveStateManager.TO_PROCESS == {
            table_id: pool
        }
        assert tbl.view().to_records() == data[:5]

        # should have flushed the process queue
        assert _PerspectiveStateManager.TO_PROCESS == {}

    def test_async_multiple_managers_queue_process(self):
        tbl = Table({
            "a": int,
            "b": float,
            "c": str
        })
        tbl2 = Table({
            "a": int,
            "b": float,
            "c": str
        })
        manager = PerspectiveManager()
        manager2 = PerspectiveManager()

        manager.host_table("tbl", tbl)
        manager2.host_table("tbl2", tbl2)

        manager._set_queue_process(TestAsync.wrapped_queue_process)
        manager2._set_queue_process(TestAsync.wrapped_queue_process)

        for i in range(5):
            tbl.update([data[i]])
            tbl2.update([data[i]])

        assert SENTINEL.get() != 0

        # flush `TO_PROCESS`
        assert tbl.view().to_records() == data[:5]

        for i in range(5):
            tbl2.update([data[i]])

    def test_async_multiple_managers_mixed_queue_process(self):
        # mutate when synchronously calling queue_process for each update
        SENTINEL_2 = AsyncSentinel(0)

        def sync_queue_process(table_id, state_manager):
            SENTINEL_2.set(SENTINEL_2.get() - 1)
            state_manager.call_process(table_id)

        tbl = Table({
            "a": int,
            "b": float,
            "c": str
        })
        tbl2 = Table({
            "a": int,
            "b": float,
            "c": str
        })
        manager = PerspectiveManager()
        manager2 = PerspectiveManager()

        manager.host_table("tbl", tbl)
        manager2.host_table("tbl2", tbl2)

        # manager uses tornado, manager2 is synchronous
        manager._set_queue_process(TestAsync.wrapped_queue_process)
        manager2._set_queue_process(sync_queue_process)

        tbl_id = tbl._table.get_id()
        tbl2_id = tbl2._table.get_id()

        for i in range(5):
            tbl.update([data[i]])
            tbl2.update([data[i]])

        assert SENTINEL.get() != 0
        assert SENTINEL_2.get() == -5

        assert tbl2_id not in _PerspectiveStateManager.TO_PROCESS

        # flush `TO_PROCESS`
        assert tbl.view().to_records() == data[:5]
        assert tbl_id not in _PerspectiveStateManager.TO_PROCESS

    def test_async_multiple_managers_delayed_process(self):
        from time import sleep
        short_delay_queue_process = partial(queue_process_async_delay,
                                            delay=0.5, loop=TestAsync.loop)
        long_delay_queue_process = partial(queue_process_async_delay,
                                           delay=1, loop=TestAsync.loop)

        tbl = Table({
            "a": int,
            "b": float,
            "c": str
        })
        tbl2 = Table({
            "a": int,
            "b": float,
            "c": str
        })
        manager = PerspectiveManager()
        manager2 = PerspectiveManager()

        manager.host_table("tbl", tbl)
        manager2.host_table("tbl2", tbl2)

        # The guarantee of `queue_process` is that eventually `_process`
        # will be called, either by user action or loop iteration. By adding
        # the delay, we can artificially queue up actions for later execution
        # and see that it's working properly.
        manager._set_queue_process(short_delay_queue_process)
        manager2._set_queue_process(long_delay_queue_process)

        tbl_id = tbl._table.get_id()
        tbl2_id = tbl2._table.get_id()

        for i in range(10):
            tbl.update([data[i]])
            tbl2.update([data[i]])

        assert SENTINEL.get() != 0

        # updates are now queued
        assert tbl_id in _PerspectiveStateManager.TO_PROCESS
        assert tbl2_id in _PerspectiveStateManager.TO_PROCESS

        # Wait for the callbacks to run - we don't call any methods
        # that would call `call_process`, but instead wait for the
        # callbacks to execute asynchronously.
        sleep(1)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from datetime import date, datetime
from .libbinding import t_dtype


def _extract_type(type, typemap):
    rval = typemap.get(type)

    if rval is None:
        raise KeyError("unsupported type: {}".format(type))

    return rval


def _dtype_to_pythontype(dtype):
    '''Returns the native Python type from a Perspective type'''
    mapping = {
        t_dtype.DTYPE_BOOL: bool,
        t_dtype.DTYPE_FLOAT32: float,
        t_dtype.DTYPE_FLOAT64: float,
        t_dtype.DTYPE_INT8: int,
        t_dtype.DTYPE_INT16: int,
        t_dtype.DTYPE_INT32: int,
        t_dtype.DTYPE_INT64: int,
        t_dtype.DTYPE_DATE: date,
        t_dtype.DTYPE_TIME: datetime,
        t_dtype.DTYPE_STR: str
    }

    return _extract_type(dtype, mapping)


def _dtype_to_str(dtype):
    '''Returns the normalized string representation of a Perspective type,
    compatible with Perspective.js.
    '''
    mapping = {
        t_dtype.DTYPE_BOOL: "boolean",
        t_dtype.DTYPE_FLOAT32: "float",
        t_dtype.DTYPE_FLOAT64: "float",
        t_dtype.DTYPE_INT8: "integer",
        t_dtype.DTYPE_INT16: "integer",
        t_dtype.DTYPE_INT32: "integer",
        t_dtype.DTYPE_INT64: "integer",
        t_dtype.DTYPE_DATE: "date",
        t_dtype.DTYPE_TIME: "datetime",
        t_dtype.DTYPE_STR: "string"
    }

    return _extract_type(dtype, mapping)


def _str_to_pythontype(typestring):
    '''Returns a Python type from the normalized string representation of a
    Perspective type, i.e. from Perspective.js
    '''
    mapping = {
        "integer": int,
        "float": float,
        "boolean": bool,
        "string": str,
        "date": date,
        "datetime": datetime
    }

    return _extract_type(typestring, mapping)



# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import os
import pyarrow as pa
from datetime import date, datetime
from perspective.table import Table

SOURCE_STREAM_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "int_float_str.arrow")
SOURCE_FILE_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "int_float_str.arrow")
PARTIAL_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "int_float_str_update.arrow")
DICT_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "dict.arrow")
DICT_UPDATE_ARROW = os.path.join(os.path.dirname(__file__), "arrow", "dict_update.arrow")

names = ["a", "b", "c", "d"]


class TestUpdateArrow(object):

    # files

    def test_update_arrow_updates_stream_file(self):
        tbl = Table({
            "a": int,
            "b": float,
            "c": str
        })

        with open(SOURCE_STREAM_ARROW, mode='rb') as file:  # b is important -> binary
            tbl.update(file.read())
            assert tbl.size() == 4
            assert tbl.schema() == {
                "a": int,
                "b": float,
                "c": str
            }

        with open(SOURCE_FILE_ARROW, mode='rb') as file:
            tbl.update(file.read())
            assert tbl.size() == 8
            assert tbl.view().to_dict() == {
                "a": [1, 2, 3, 4] * 2,
                "b": [1.5, 2.5, 3.5, 4.5] * 2,
                "c": ["a", "b", "c", "d"] * 2
            }

    def test_update_arrow_partial_updates_file(self):
        tbl = Table({
            "a": int,
            "b": float,
            "c": str
        }, index="a")

        with open(SOURCE_STREAM_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 4

        with open(PARTIAL_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 4
            assert tbl.view().to_dict() == {
                "a": [1, 2, 3, 4],
                "b": [100.5, 2.5, 3.5, 400.5],
                "c": ["x", "b", "c", "y"]
            }

    def test_update_arrow_updates_dict_file(self):
        tbl = Table({
            "a": str,
            "b": str
        })

        with open(DICT_ARROW, mode='rb') as src:
            tbl.update(src.read())
            assert tbl.size() == 5

        with open(DICT_UPDATE_ARROW, mode='rb') as partial:
            tbl.update(partial.read())
            assert tbl.size() == 8
            assert tbl.view().to_dict() == {
                "a": ["abc", "def", "def", None, "abc", None, "update1", "update2"],
                "b": ["klm", "hij", None, "hij", "klm", "update3", None, "update4"]
            }

    # streams

    def test_update_arrow_updates_int_stream(self, util):
        data = [list(range(10)) for i in range(4)]
        arrow_data = util.make_arrow(names, data)
        tbl = Table({
            "a": int,
            "b": int,
            "c": int,
            "d": int
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0],
            "b": data[1],
            "c": data[2],
            "d": data[3]
        }

    def test_update_arrow_updates_float_stream(self, util):
        data = [
            [i for i in range(10)],
            [i * 1.5 for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a", "b"], data)
        tbl = Table({
            "a": int,
            "b": float,
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0],
            "b": data[1]
        }

    def test_update_arrow_updates_decimal128_stream(self, util):
        data = [
            [i * 1000000000 for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.decimal128(10)])
        tbl = Table({
            "a": int
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0]
        }

    def test_update_arrow_updates_bool_stream(self, util):
        data = [
            [True if i % 2 == 0 else False for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data)
        tbl = Table({
            "a": bool
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0]
        }

    def test_update_arrow_updates_date32_stream(self, util):
        data = [
            [date(2019, 2, i) for i in range(1, 11)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.date32()])
        tbl = Table({
            "a": date
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 2, i) for i in range(1, 11)]
        }

    def test_update_arrow_updates_date64_stream(self, util):
        data = [
            [date(2019, 2, i) for i in range(1, 11)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.date64()])
        tbl = Table({
            "a": date
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": [datetime(2019, 2, i) for i in range(1, 11)]
        }

    def test_update_arrow_updates_timestamp_all_formats_stream(self, util):
        data = [
            [datetime(2019, 2, i, 9) for i in range(1, 11)],
            [datetime(2019, 2, i, 10) for i in range(1, 11)],
            [datetime(2019, 2, i, 11) for i in range(1, 11)],
            [datetime(2019, 2, i, 12) for i in range(1, 11)]
        ]
        arrow_data = util.make_arrow(
            names, data, types=[
                pa.timestamp("s"),
                pa.timestamp("ms"),
                pa.timestamp("us"),
                pa.timestamp("ns"),
            ]
        )
        tbl = Table({
            "a": datetime,
            "b": datetime,
            "c": datetime,
            "d": datetime
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0],
            "b": data[1],
            "c": data[2],
            "d": data[3]
        }

    def test_update_arrow_updates_string_stream(self, util):
        data = [
            [str(i) for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.string()])
        tbl = Table({
            "a": str
        })
        tbl.update(arrow_data)
        assert tbl.size() == 10
        assert tbl.view().to_dict() == {
            "a": data[0]
        }

    def test_update_arrow_updates_dictionary_stream(self, util):
        data = [
            ([0, 1, 1, None], ["a", "b"]),
            ([0, 1, None, 2], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data)
        tbl = Table({
            "a": str,
            "b": str
        })
        tbl.update(arrow_data)

        assert tbl.size() == 4
        assert tbl.view().to_dict() == {
            "a": ["a", "b", "b", None],
            "b": ["x", "y", None, "z"]
        }

    def test_update_arrow_arbitary_order(self, util):
        data = [[1, 2, 3, 4],
                ["a", "b", "c", "d"],
                [1, 2, 3, 4],
                ["a", "b", "c", "d"]]
        update_data = [[5, 6], ["e", "f"], [5, 6], ["e", "f"]]
        arrow = util.make_arrow(["a", "b", "c", "d"], data)
        update_arrow = util.make_arrow(["c", "b", "a", "d"], update_data)
        tbl = Table(arrow)
        assert tbl.schema() == {
            "a": int,
            "b": str,
            "c": int,
            "d": str
        }
        tbl.update(update_arrow)
        assert tbl.size() == 6
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3, 4, 5, 6],
            "b": ["a", "b", "c", "d", "e", "f"],
            "c": [1, 2, 3, 4, 5, 6],
            "d": ["a", "b", "c", "d", "e", "f"]
        }

    # append

    def test_update_arrow_updates_append_int_stream(self, util):
        data = [list(range(10)) for i in range(4)]
        arrow_data = util.make_arrow(names, data)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0],
            "b": data[1] + data[1],
            "c": data[2] + data[2],
            "d": data[3] + data[3]
        }

    def test_update_arrow_updates_append_float_stream(self, util):
        data = [
            [i for i in range(10)],
            [i * 1.5 for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a", "b"], data)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0],
            "b": data[1] + data[1]
        }

    def test_update_arrow_updates_append_decimal_stream(self, util):
        data = [
            [i * 1000 for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.decimal128(4)])
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0]
        }

    def test_update_arrow_updates_append_bool_stream(self, util):
        data = [
            [True if i % 2 == 0 else False for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0]
        }

    def test_update_arrow_updates_append_date32_stream(self, util):
        data = [
            [date(2019, 2, i) for i in range(1, 11)]
        ]
        out_data = [datetime(2019, 2, i) for i in range(1, 11)]
        arrow_data = util.make_arrow(["a"], data, types=[pa.date32()])
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": out_data + out_data
        }

    def test_update_arrow_updates_append_date64_stream(self, util):
        data = [
            [date(2019, 2, i) for i in range(1, 11)]
        ]
        out_data = [datetime(2019, 2, i) for i in range(1, 11)]
        arrow_data = util.make_arrow(["a"], data, types=[pa.date64()])
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": out_data + out_data
        }

    def test_update_arrow_updates_append_timestamp_all_formats_stream(self, util):
        data = [
            [datetime(2019, 2, i, 9) for i in range(1, 11)],
            [datetime(2019, 2, i, 10) for i in range(1, 11)],
            [datetime(2019, 2, i, 11) for i in range(1, 11)],
            [datetime(2019, 2, i, 12) for i in range(1, 11)]
        ]
        arrow_data = util.make_arrow(
            names, data, types=[
                pa.timestamp("s"),
                pa.timestamp("ms"),
                pa.timestamp("us"),
                pa.timestamp("ns"),
            ]
        )
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0],
            "b": data[1] + data[1],
            "c": data[2] + data[2],
            "d": data[3] + data[3],
        }

    def test_update_arrow_updates_append_string_stream(self, util):
        data = [
            [str(i) for i in range(10)]
        ]
        arrow_data = util.make_arrow(["a"], data, types=[pa.string()])
        tbl = Table(arrow_data)
        tbl.update(arrow_data)
        assert tbl.size() == 20
        assert tbl.view().to_dict() == {
            "a": data[0] + data[0]
        }

    def test_update_arrow_updates_append_dictionary_stream(self, util):
        data = [
            ([0, 1, 1, None], ["a", "b"]),
            ([0, 1, None, 2], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)

        assert tbl.size() == 8
        assert tbl.view().to_dict() == {
            "a": ["a", "b", "b", None, "a", "b", "b", None],
            "b": ["x", "y", None, "z", "x", "y", None, "z"]
        }

    def test_update_arrow_updates_append_dictionary_stream_legacy(self, util):
        data = [
            ([0, 1, 1, None], ["a", "b"]),
            ([0, 1, None, 2], ["x", "y", "z"])
        ]
        arrow_data = util.make_dictionary_arrow(["a", "b"], data, legacy=True)
        tbl = Table(arrow_data)
        tbl.update(arrow_data)

        assert tbl.size() == 8
        assert tbl.view().to_dict() == {
            "a": ["a", "b", "b", None, "a", "b", "b", None],
            "b": ["x", "y", None, "z", "x", "y", None, "z"]
        }

    # indexed

    def test_update_arrow_partial_indexed(self, util):
        data = [[1, 2, 3, 4], ["a", "b", "c", "d"]]
        update_data = [[2, 4], ["x", "y"]]
        arrow = util.make_arrow(["a", "b"], data)
        update_arrow = util.make_arrow(["a", "b"], update_data)
        tbl = Table(arrow, index="a")
        assert tbl.schema() == {
            "a": int,
            "b": str
        }
        tbl.update(update_arrow)
        assert tbl.size() == 4
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3, 4],
            "b": ["a", "x", "c", "y"]
        }

    # update specific columns

    def test_update_arrow_specific_column(self, util):
        data = [[1, 2, 3, 4], ["a", "b", "c", "d"]]
        update_data = [[2, 3, 4]]
        arrow = util.make_arrow(["a", "b"], data)
        update_arrow = util.make_arrow(["a"], update_data)
        tbl = Table(arrow)
        assert tbl.schema() == {
            "a": int,
            "b": str
        }
        tbl.update(update_arrow)
        assert tbl.size() == 7
        assert tbl.view().to_dict() == {
            "a": [1, 2, 3, 4, 2, 3, 4],
            "b": ["a", "b", "c", "d", None, None, None]
        }

    # try to fuzz column order

    def test_update_arrow_column_order_str(self, util):
        # use str so it doesn't get promoted
        data = [["a", "b", "c"] for i in range(10)]
        names = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
        names_scrambled = names[::-1]
        arrow = util.make_arrow(names_scrambled, data)
        tbl = Table({name: str for name in names})
        tbl.update(arrow)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            name: data[0] for name in names
        }

    def test_update_arrow_column_order_int(self, util):
        data = [[1, 2, 3] for i in range(10)]
        names = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
        names_scrambled = names[::-1]
        arrow = util.make_arrow(names_scrambled, data)
        tbl = Table({name: int for name in names})
        tbl.update(arrow)
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            name: data[0] for name in names
        }



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from ._version import __version__  # noqa: F401
from .exception import PerspectiveError  # noqa: F401
from .aggregate import Aggregate  # noqa: F401
from .filters import ALL_FILTERS  # noqa: F401
from .plugin import Plugin  # noqa: F401
from .sort import Sort  # noqa: F401



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from pytest import raises
from perspective import PerspectiveError, PerspectiveViewer,\
                        PerspectiveWidget, Plugin


class TestPlugin:

    def test_plugin_widget_load(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data, plugin=Plugin.XBAR)
        assert widget.plugin == "x_bar"

    def test_plugin_widget_setattr(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data)
        widget.plugin = Plugin.XBAR
        assert widget.plugin == "x_bar"

    def test_plugin_widget_load_invalid(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        with raises(PerspectiveError):
            PerspectiveWidget(data, plugin="?")

    def test_plugin_widget_setattr_invalid(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data) 
        with raises(PerspectiveError):
            widget.plugin = "?"

    def test_plugin_widget_init_all(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        for plugin in Plugin:
            widget = PerspectiveWidget(data, plugin=plugin)
            assert widget.plugin == plugin.value

    def test_plugin_widget_set_all(self):
        data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
        widget = PerspectiveWidget(data)
        for plugin in Plugin:
            widget.plugin = plugin
            assert widget.plugin == plugin.value

    def test_plugin_viewer_load(self):
        viewer = PerspectiveViewer(plugin=Plugin.XBAR)
        assert viewer.plugin == "x_bar"

    def test_plugin_viewer_setattr(self):
        viewer = PerspectiveViewer()
        viewer.plugin = Plugin.XBAR
        assert viewer.plugin == "x_bar"

    def test_plugin_viewer_init_all(self):
        for plugin in Plugin:
            viewer = PerspectiveViewer(plugin=plugin)
            assert viewer.plugin == plugin.value

    def test_plugin_viewer_set_all(self):
        viewer = PerspectiveViewer()
        for plugin in Plugin:
            viewer.plugin = plugin
            assert viewer.plugin == plugin.value



import six
import inspect
from ..utils import _either_type


class Node(object):
    '''Class to represent an operation that is lazy'''
    _id_ref = 0

    def __init__(self,
                 value=None,
                 name="?",
                 derived=False,
                 readonly=False,
                 nullable=False,
                 callable=None,
                 callable_args=None,
                 callable_kwargs=None,
                 callable_is_method=False,
                 always_dirty=False,
                 trace=False,
                 **kwargs
                 ):
        '''Construct a new lazy node, wrapping a callable or a value

        Args:
            name (str): name to use to represent the node
            derived (bool):
            readonly (bool): whether a node is settable
            nullable (bool): whether a node can have value None
            value (any): initial value of the node
            callable (callable): function or other callable that the node is wrapping
            callable_args (tuple): args for the wrapped callable
            callable_kwargs (dict): kwargs for the wrapped callable
            callable_is_method (bool): is the callable a method of an object
            always_dirty (bool): node should not be lazy - always access underlying value
            trace (bool): trace when the node is called
        '''
        # Instances get an id but one id tracker for all nodes so we can
        # uniquely identify them
        # TODO different scheme
        self._id = Node._id_ref
        Node._id_ref += 1

        # Every node gets a name so it can be uniquely identified in the graph
        self._name = '{}#{}'.format(name or self.__class__.__name__, self._id)

        if isinstance(value, Node):
            raise Exception('Cannot set value to be itself a node')

        # Graphviz shape
        self._graphvizshape = kwargs.get('graphvizshape', 'box')  # default is box instead of ellipse
        # because all lazy nodes are i/o nodes
        # by default

        # if using dagre-d3, this will be set
        self._dd3g = None

        # starting value
        self._value = value

        # can be set
        self._readonly = readonly

        # trace calls/executions
        self._trace = trace

        # callable and args
        if not inspect.isgeneratorfunction(callable):
            self._callable = callable
        else:
            def _callable(gen=callable(*(callable_args or []), **(callable_kwargs or {}))):
                try:
                    return next(gen)
                except StopIteration:
                    self._always_dirty = False
                    self._dirty = False
                    return self._value
            self._callable = _callable
        self._callable_args = self._transform_args(callable_args or [])
        self._callable_kwargs = self._transform_kwargs(callable_kwargs or {})
        self._callable_is_method = callable_is_method

        # if always dirty, always reevaluate
        self._always_dirty = always_dirty or not(self._callable is None)

        # parent nodes in graph
        self._parents = []

        # self reference for method calls
        self._self_reference = self

        # cache node operations that have already been done
        self._node_op_cache = {}

        # dependencies can be nodes
        if self._callable:
            self._callable._node_wrapper = None  # not known until program start
            self._dependencies = {self._callable: (self._callable_args, self._callable_kwargs)}
        else:
            self._dependencies = {}

        # if derived node, default to dirty to start
        if derived:
            self._dirty = True
        else:
            self._dirty = False

    def _get_dirty(self):
        return self._is_dirty

    def _set_dirty(self, val):
        if val:
            self._reddd3g()
        else:
            self._whited3g()
        self._is_dirty = val

    _dirty = property(_get_dirty, _set_dirty)

    def _name_no_id(self):
        return self._name.rsplit('#', 1)[0]

    def _transform_args(self, args):
        return args

    def _transform_kwargs(self, kwargs):
        return kwargs

    def _with_self(self, other_self):
        self._self_reference = other_self
        return self

    def _greendd3g(self):
        if self._dd3g:
            self._dd3g.setNode(self._name, tooltip=str(self.value()), style='fill: #0f0')

    def _yellowdd3g(self):
        if self._dd3g:
            self._dd3g.setNode(self._name, tooltip=str(self.value()), style='fill: #ff0')

    def _reddd3g(self):
        if self._dd3g:
            self._dd3g.setNode(self._name, tooltip=str(self.value()), style='fill: #f00')

    def _whited3g(self):
        if self._dd3g:
            self._dd3g.setNode(self._name, tooltip=str(self.value()), style='fill: #fff')

    def _compute_from_dependencies(self):
        if self._dependencies:
            self._greendd3g()
            for deps in six.itervalues(self._dependencies):
                # recompute args
                for arg in deps[0]:
                    arg._recompute()

                    # Set yourself as parent
                    if self not in arg._parents:
                        arg._parents.append(self)

                # recompute kwargs
                for kwarg in six.itervalues(deps[1]):
                    kwarg._recompute()

                    # Set yourself as parent
                    if self not in kwarg._parents:
                        kwarg._parents.append(self)

            k = list(self._dependencies.keys())[0]

            if self._callable_is_method:
                new_value = k(self._self_reference, *self._dependencies[k][0], **self._dependencies[k][1])
            else:
                new_value = k(*self._dependencies[k][0], **self._dependencies[k][1])

            if isinstance(new_value, Node):
                k._node_wrapper = new_value
                new_value = new_value()  # get value

            if isinstance(new_value, Node):
                raise Exception('Value should not itself be a node!')

            if self._trace:
                if new_value != self._value:
                    print('recomputing: %s#%d' % (self._name, id(self)))

            self._value = new_value

        self._whited3g()
        return self._value

    def _subtree_dirty(self):
        for call, deps in six.iteritems(self._dependencies):
            # callable node
            if hasattr(call, '_node_wrapper') and \
               call._node_wrapper is not None:
                if call._node_wrapper.isDirty():
                    # CRITICAL
                    # always set self to be dirty if subtree is dirty
                    self._dirty = True
                    return True

            # check args
            for arg in deps[0]:
                if arg.isDirty():
                    # CRITICAL
                    # always set self to be dirty if subtree is dirty
                    self._dirty = True
                    return True

            # check kwargs
            for kwarg in six.itervalues(deps[1]):
                if kwarg.isDirty():
                    # CRITICAL
                    # always set self to be dirty if subtree is dirty
                    self._dirty = True
                    return True
        return False

    def isDirty(self):
        self._dirty = self._dirty or self._subtree_dirty() or self._always_dirty
        return self._dirty

    def _recompute(self):
        self.isDirty()
        if self._dirty:
            if self._parents:
                for parent in self._parents:
                    # let your parents know you were dirty!
                    parent._dirty = True
            self._value = self._compute_from_dependencies()
        self._dirty = False

    def _gennode(self, name, foo, foo_args, trace=False, **kwargs):
        if name not in self._node_op_cache:
            self._node_op_cache[name] = \
                Node(name=name,
                     derived=True,
                     callable=foo,
                     callable_args=foo_args,
                     trace=trace,
                     **kwargs)
        return self._node_op_cache[name]

    def _tonode(self, other, trace=False):
        if isinstance(other, Node):
            return other
        if str(other) not in self._node_op_cache:
            self._node_op_cache[str(other)] = \
                Node(name='var(' + str(other)[:5] + ')',
                     derived=True,
                     value=other,
                     trace=trace)
        return self._node_op_cache[str(other)]

    def setValue(self, value):
        if value != self._value:
            self._value = value  # leave for dagre
            self._dirty = True
        self._value = value

    def set(self, *args, **kwargs):
        for k, v in six.iteritems(kwargs):
            _set = False
            for deps in six.itervalues(self._dependencies):
                # try to set args
                for arg in deps[0]:
                    if arg._name_no_id() == k:
                        arg._dirty = (arg._value != v)
                        arg._value = v
                        _set = True
                        break

                if _set:
                    continue

                # try to set kwargs
                for kwarg in six.itervalues(deps[1]):
                    if kwarg._name_no_id() == k:
                        kwarg._dirty = (kwarg._value != v)
                        kwarg._value = v
                        _set = True
                        break

    def value(self):
        return self._value

    def __call__(self):
        self._recompute()
        return self.value()

    def __repr__(self):
        return self._name


@_either_type
def node(meth, memoize=True, trace=False):
    argspec = inspect.getfullargspec(meth)
    # args = argspec.args
    # varargs = argspec.varargs
    # varkw = argspec.varkw
    # defaults = argspec.defaults
    # kwonlyargs = argspec.kwonlyargs
    # kwonlydefaults = args.kwonlydefaults

    if argspec.varargs:
        raise Exception('varargs not supported yet!')
    if argspec.varkw:
        raise Exception('varargs not supported yet!')

    node_args = []
    node_kwargs = {}
    is_method = False

    for i, arg in enumerate(argspec.args):
        if arg == 'self':
            # TODO
            is_method = True
            continue

        if (is_method and len(argspec.defaults or []) >= i) or \
           (not is_method and len(argspec.defaults or []) > i):
            value = argspec.defaults[0]
            nullable = True
        else:
            value = None
            nullable = False

        node_args.append(Node(name=arg,
                              derived=True,
                              readonly=False,
                              nullable=nullable,
                              value=value,
                              trace=trace))

    for k, v in six.iteritems(argspec.kwonlydefaults or {}):
        node_kwargs[k] = Node(name=k,
                              derived=True,
                              readonly=False,
                              nullable=True,
                              value=v,
                              trace=trace)

    def meth_wrapper(self, *args, **kwargs):
        if len(args) > len(node_args):
            raise Exception('Missing args (call or preprocessing error has occurred)')

        if len(kwargs) > len(node_kwargs):
            raise Exception('Missing kwargs (call or preprocessing error has occurred)')

        if is_method:
            val = meth(self, *(arg.value() for arg in args), **kwargs)
        else:
            val = meth(*(arg.value() for arg in args), **kwargs)
        return val

    new_node = Node(name=meth.__name__,
                    derived=True,
                    callable=meth_wrapper,
                    callable_args=node_args,
                    callable_kwargs=node_kwargs,
                    callable_is_method=is_method,
                    always_dirty=not memoize,
                    trace=trace)

    if is_method:
        ret = lambda self, *args, **kwargs: new_node._with_self(self)  # noqa: E731
    else:
        ret = lambda *args, **kwargs: new_node  # noqa: E731

    ret._node_wrapper = new_node
    return ret



import tributary.streaming as ts


class TestConst:
    def test_const_1(self):
        t = ts.Const(value=1, count=1)
        assert ts.run(t) == [1]

    def test_const_2(self):
        t = ts.Const(value=1, count=5)
        assert ts.run(t) == [1, 1, 1, 1, 1]


class TestTimer:
    def test_timer(self):
        val = 0

        def foo():
            nonlocal val
            val += 1
            return val

        t = ts.Timer(foo, count=5)
        assert ts.run(t) == [1, 2, 3, 4, 5]

        t = ts.Timer(foo, count=5)

    def test_timer_delay(self):
        val = 0

        def foo():
            nonlocal val
            val += 1
            return val

        t = ts.Timer(foo, count=5, interval=0.1)
        assert ts.run(t) == [1, 2, 3, 4, 5]

        t = ts.Timer(foo, count=5)

    def test_timer_generator(self):
        def foo():
            yield 1
            yield 2
            yield 3
            yield 4
            yield 5

        t = ts.Timer(foo)
        assert ts.run(t) == [1]

        t = ts.Timer(foo, count=3)
        assert ts.run(t) == [1, 2, 3]

        t = ts.Timer(foo, count=5)
        assert ts.run(t) == [1, 2, 3, 4, 5]

        t = ts.Timer(foo, count=6)
        assert ts.run(t) == [1, 2, 3, 4, 5]

    def test_timer_generator_delay(self):
        def foo():
            yield 1
            yield 2
            yield 3
            yield 4
            yield 5

        t = ts.Timer(foo, interval=0.1)
        assert ts.run(t) == [1]

        t = ts.Timer(foo, count=3, interval=0.1)
        assert ts.run(t) == [1, 2, 3]

        t = ts.Timer(foo, count=5, interval=0.1)
        assert ts.run(t) == [1, 2, 3, 4, 5]

        t = ts.Timer(foo, count=6, interval=0.1)
        assert ts.run(t) == [1, 2, 3, 4, 5]


class TestFoo:
    def test_timer(self):
        val = 0

        def foo():
            nonlocal val
            val += 1
            return val

        t = ts.Timer(foo, count=5)
        assert ts.run(t) == [1, 2, 3, 4, 5]

        t = ts.Timer(foo, count=5)

    def test_timer_delay(self):
        val = 0

        def foo():
            nonlocal val
            val += 1
            return val

        t = ts.Timer(foo, count=5, interval=0.1)
        assert ts.run(t) == [1, 2, 3, 4, 5]

        t = ts.Timer(foo, count=5)

    def test_foo_generator(self):
        def foo():
            yield 1
            yield 2
            yield 3
            yield 4
            yield 5

        t = ts.Foo(foo)
        assert ts.run(t) == [1, 2, 3, 4, 5]



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from .libpsp import *  # noqa: F401, F403
from .core import *  # noqa: F401, F403
from .core._version import __version__  # noqa: F401
from .widget import *  # noqa: F401, F403

try:
    from .node import *  # noqa: F401, F403
except ImportError:
    pass



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import numpy as np
import pandas as pd


def _parse_datetime_index(index):
    '''Given an instance of `pandas.DatetimeIndex`, parse its `freq` and
    return a `numpy.dtype` that corresponds to the unit it should be parsed in.

    Because `pandas.DataFrame`s cannot store datetimes in anything other than
    `datetime64[ns]`, we need to examine the `DatetimeIndex` itself to
    understand what unit it needs to be parsed as.

    Args:
        index (pandas.DatetimeIndex)

    Returns:
        `numpy.dtype`: a datetime64 dtype with the correct units depending on
            `index.freq`.
    '''
    if index.freq is None:
        return np.dtype("datetime64[ns]")

    freq = str(index.freq).lower()
    new_type = None

    if any(s in freq for s in ["businessday", "day"]) or freq == "sm" or freq == "sms":
        # days
        new_type = "D"
    elif freq == "w" or "week" in freq:
        # weeks
        new_type = "W"
    elif any(s in freq for s in ["month", "quarter"]):
        # months
        new_type = "M"
    elif "year" in freq or freq == "a":
        new_type = "Y"
    else:
        # default to datetime
        new_type = "ns"

    return np.dtype("datetime64[{0}]".format(new_type))


def deconstruct_pandas(data):
    '''Given a dataframe, flatten it by resetting the index and memoizing the
    pivots that were applied.

    Args:
        data (pandas.dataframe): a Pandas DataFrame to parse

    Returns:
        (pandas.DataFrame, dict): a Pandas DataFrame and a dictionary containing
            optional members `columns`, `row_pivots`, and `column_pivots`.
    '''
    kwargs = {}

    # level unstacking
    if isinstance(data, pd.DataFrame) and isinstance(data.columns, pd.MultiIndex):
        data = pd.DataFrame(data.unstack())
        columns = list(x for x in data.index.names if x)
        kwargs['column_pivots'] = list(x for x in data.index.names if x)
        kwargs['row_pivots'] = []
        orig_columns = [' ' for _ in data.columns.tolist()]

        # deal with indexes
        if len(columns) < len(data.index.names):
            for i in range(len(data.index.names) - len(columns)):
                if i == 0:
                    columns.append('index')
                    kwargs['row_pivots'].append('index')
                else:
                    columns.append('index-{}'.format(i))
                    kwargs['row_pivots'].append('index-{}'.format(i))

        # all columns in index
        columns += orig_columns
        data.reset_index(inplace=True)
        data.columns = columns

        # use these columns
        kwargs['columns'] = orig_columns

    # Decompose Period index to timestamps
    if isinstance(data.index, pd.PeriodIndex):
        data.index = data.index.to_timestamp()

    if isinstance(data.index, pd.MultiIndex):
        kwargs['row_pivots'] = list(data.index.names)
        kwargs['columns'] = data.columns.tolist()

    if isinstance(data, pd.Series) or 'index' not in map(lambda x: str(x).lower(), data.columns):
        flattened = data.reset_index()

        if isinstance(data, pd.Series):
            # preserve name from series
            flattened.name = data.name

            # make sure all columns are strings
            flattened.columns = [str(c) for c in flattened.columns]

        data = flattened

    return data, kwargs



from IPython.display import display
from ..base import Node, _gen_node

_OUTPUT_GRAPHVIZSHAPE = "box"


def Print(node, text=''):
    def foo(val):
        print(text + str(val))
        return val

    node = _gen_node(node)
    ret = Node(foo=foo, foo_kwargs=None, name='Print', inputs=1, graphvizshape=_OUTPUT_GRAPHVIZSHAPE)

    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret


def Graph(node):
    if not node._upstream:
        # leaf node
        return {node: []}
    return {node: [_.graph() for _ in node._upstream]}


def PPrint(node, level=0):
    ret = '    ' * (level - 1) if level else ''

    if not node._upstream:
        # leaf node
        return ret + '  \\  ' + repr(node)
    return '    ' * level + repr(node) + '\n' + '\n'.join(_.pprint(level + 1) for _ in node._upstream)


def GraphViz(node):
    d = Graph(node)

    from graphviz import Digraph
    dot = Digraph("Graph", strict=False)
    dot.format = 'png'

    def rec(nodes, parent):
        for d in nodes:
            if not isinstance(d, dict):
                dot.node(d, shape=d._graphvizshape)
                dot.edge(d, parent)

            else:
                for k in d:
                    dot.node(k._name, shape=k._graphvizshape)
                    rec(d[k], k)
                    dot.edge(k._name, parent._name)

    for k in d:
        dot.node(k._name, shape=k._graphvizshape)
        rec(d[k], k)

    return dot


def Dagre(node):
    import ipydagred3 as dd3
    G = dd3.Graph()
    d = Graph(node)

    def rec(nodes, parent):
        for d in nodes:
            if not isinstance(d, dict):
                d._dd3g = G
                G.setNode(d._name, shape="rect" if d._graphvizshape == "box" else d._graphvizshape)
                G.setEdge(d._name, parent)
            else:
                for k in d:
                    k._dd3g = G
                    G.setNode(k._name, shape="rect" if k._graphvizshape == "box" else k._graphvizshape)
                    G.setEdge(k._name, parent._name)
                    rec(d[k], k)

    for k in d:
        k._dd3g = G
        G.setNode(k._name, shape="rect" if k._graphvizshape == "box" else k._graphvizshape)
        rec(d[k], k)

    graph = dd3.DagreD3Widget(graph=G)
    return graph


def Perspective(node, text='', **psp_kwargs):
    psp_kwargs = psp_kwargs or {}
    from perspective import PerspectiveWidget
    p = PerspectiveWidget(psp_kwargs.pop('schema', []), **psp_kwargs)

    def foo(val):
        p.update(val)
        return val

    node = _gen_node(node)
    ret = Node(foo=foo, foo_kwargs=None, name='Perspective', inputs=1, graphvizshape=_OUTPUT_GRAPHVIZSHAPE)

    display(p)
    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret


Node.graph = Graph
Node.pprint = PPrint
Node.graphviz = GraphViz
Node.dagre = Dagre
Node.print = Print
Node.perspective = Perspective



import tributary.streaming as ts


class TestOutput:
    def test_print(self):
        val = 0

        def foo():
            nonlocal val
            val += 1
            return val

        assert ts.run(ts.Print(ts.Timer(foo, count=2))) == [1, 2]

    def test_print_generator(self):
        def foo():
            yield 1
            yield 2
            yield 3
            yield 4
            yield 5

        assert ts.run(ts.Print(ts.Timer(foo, count=2))) == [1, 2]

    def test_graphviz(self):
        def foo():
            yield 1

        assert ts.GraphViz(ts.Print(ts.Timer(foo, count=2)))

    def test_dagre(self):
        def foo():
            yield 1

        assert ts.Dagre(ts.Print(ts.Timer(foo, count=2)))



import math
import numpy as np
from ..base import Node


_INPUT_GRAPHVIZSHAPE = "box"


def _gen():
    S = 100
    T = 252
    mu = 0.25
    vol = 0.5

    returns = np.random.normal(mu / T, vol / math.sqrt(T), T) + 1
    _list = returns.cumprod() * S
    return _list


class Timer(Node):
    '''Streaming wrapper to periodically call a callable `count` times
       with a delay of `interval` in between

    Arguments:
        foo (callable): callable to call
        foo_kwargs (dict): kwargs for callable
        count (int): number of times to call, 0 means infinite (or until generator is complete)
        interval (int/float): minimum delay between calls (can be more due to async scheduling)
    '''

    def __init__(self, foo, foo_kwargs=None, count=1, interval=0):
        super().__init__(foo=foo,
                         foo_kwargs=foo_kwargs,
                         name='Timer[{}]'.format(foo.__name__),
                         inputs=0,
                         execution_max=count,
                         delay_interval=interval,
                         graphvizshape=_INPUT_GRAPHVIZSHAPE)


class Const(Timer):
    '''Streaming wrapper to return a scalar value

    Arguments:
        value (any): value to return
        count (int): number of times to call, 0 means infinite
    '''

    def __init__(self, value, count=0):
        super().__init__(foo=lambda: value, count=count, interval=0)
        self._name = 'Const[{}]'.format(value)


class Curve(Timer):
    '''Streaming wrapper to output a series of values

    Arguments:
        value (any): value to unroll and return
    '''

    def __init__(self, value):
        def foo(curve=value):
            for v in curve:
                yield v
        super().__init__(foo=foo, count=0)
        self._name = 'Curve[{}]'.format(len(value))


class Foo(Timer):
    '''Streaming wrapper to periodically call a function `count` times
       with a delay of `interval` in between

    Arguments:
        foo (callable): callable to call
        foo_kwargs (dict): kwargs for callable
        count (int): number of times to call, 0 means infinite (or until generator is complete)
        interval (int/float): minimum delay between calls (can be more due to async scheduling)
    '''

    def __init__(self, foo, foo_kwargs=None, count=0, interval=0):
        super().__init__(foo=foo, foo_kwargs=foo_kwargs, count=count, interval=interval)
        self._name = 'Foo[{}]'.format(foo.__name__)


class Random(Foo):
    '''Yield a random dictionary of data

    Args:
        count (int): number of elements to yield
        interval (float): interval to wait between yields
    '''

    def __init__(self, count=10, interval=0.1):
        def _random(count=count, interval=interval):
            step = 0
            while step < count:
                x = {y: _gen() for y in ('A', 'B', 'C', 'D')}
                for i in range(len(x['A'])):
                    if step >= count:
                        break
                    yield {'A': x['A'][i],
                           'B': x['B'][i],
                           'C': x['C'][i],
                           'D': x['D'][i]}
                    step += 1
        super().__init__(foo=_random, count=count, interval=interval)
        self._name = 'Random'



import os
import os.path
import tributary.streaming as ts


class TestHttp:
    def test_http(self):
        out = ts.Print(ts.HTTP(url='https://google.com'))
        ret = ts.run(out)
        print(ret)
        assert len(ret) == 1
